# -*- coding: utf-8 -*-

# Copyright (c) 2025, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.


def normalize(value):
    """Returns the string with decimal separators normalized."""
    return value.replace(",", ".")



# -*- coding: utf-8 -*-

# Copyright (c) 2025, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import datetime
import pickle
import unittest

from aniso8601.utcoffset import UTCOffset


class TestUTCOffset(unittest.TestCase):
    def test_pickle(self):
        # Make sure timezone objects are pickleable
        testutcoffset = UTCOffset(name="UTC", minutes=0)

        utcoffsetpickle = pickle.dumps(testutcoffset)

        resultutcoffset = pickle.loads(utcoffsetpickle)

        self.assertEqual(resultutcoffset._name, testutcoffset._name)
        self.assertEqual(resultutcoffset._utcdelta, testutcoffset._utcdelta)

    def test_repr(self):
        self.assertEqual(str(UTCOffset(minutes=0)), "+0:00:00 UTC")

        self.assertEqual(str(UTCOffset(minutes=60)), "+1:00:00 UTC")

        self.assertEqual(str(UTCOffset(minutes=-60)), "-1:00:00 UTC")

        self.assertEqual(str(UTCOffset(minutes=12)), "+0:12:00 UTC")

        self.assertEqual(str(UTCOffset(minutes=-12)), "-0:12:00 UTC")

        self.assertEqual(str(UTCOffset(minutes=83)), "+1:23:00 UTC")

        self.assertEqual(str(UTCOffset(minutes=-83)), "-1:23:00 UTC")

        self.assertEqual(str(UTCOffset(minutes=1440)), "+1 day, 0:00:00 UTC")

        self.assertEqual(str(UTCOffset(minutes=-1440)), "-1 day, 0:00:00 UTC")

        self.assertEqual(str(UTCOffset(minutes=2967)), "+2 days, 1:27:00 UTC")

        self.assertEqual(str(UTCOffset(minutes=-2967)), "-2 days, 1:27:00 UTC")

    def test_dst(self):
        tzinfoobject = UTCOffset(minutes=240)
        # This would raise ISOFormatError or a TypeError if dst info is invalid
        result = datetime.datetime.now(tzinfoobject)
        # Hacky way to make sure the tzinfo is what we'd expect
        self.assertEqual(result.tzinfo.utcoffset(None), datetime.timedelta(hours=4))



# -*- coding: utf-8 -*-

# Copyright (c) 2025, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

from aniso8601.builders.python import PythonTimeBuilder
from aniso8601.compat import is_string
from aniso8601.exceptions import ISOFormatError


def parse_timezone(tzstr, builder=PythonTimeBuilder):
    # tzstr can be Z, ±hh:mm, ±hhmm, ±hh
    if is_string(tzstr) is False:
        raise ValueError("Time zone must be string.")

    if len(tzstr) == 1 and tzstr[0] == "Z":
        return builder.build_timezone(negative=False, Z=True, name=tzstr)

    if len(tzstr) == 6:
        # ±hh:mm
        hourstr = tzstr[1:3]
        minutestr = tzstr[4:6]

        if tzstr[0] == "-" and hourstr == "00" and minutestr == "00":
            raise ISOFormatError("Negative ISO 8601 time offset must not be 0.")
    elif len(tzstr) == 5:
        # ±hhmm
        hourstr = tzstr[1:3]
        minutestr = tzstr[3:5]

        if tzstr[0] == "-" and hourstr == "00" and minutestr == "00":
            raise ISOFormatError("Negative ISO 8601 time offset must not be 0.")
    elif len(tzstr) == 3:
        # ±hh
        hourstr = tzstr[1:3]
        minutestr = None

        if tzstr[0] == "-" and hourstr == "00":
            raise ISOFormatError("Negative ISO 8601 time offset must not be 0.")
    else:
        raise ISOFormatError('"{0}" is not a valid ISO 8601 time offset.'.format(tzstr))

    for componentstr in [hourstr, minutestr]:
        if componentstr is not None:
            if componentstr.isdigit() is False:
                raise ISOFormatError(
                    '"{0}" is not a valid ISO 8601 time offset.'.format(tzstr)
                )

    if tzstr[0] == "+":
        return builder.build_timezone(
            negative=False, hh=hourstr, mm=minutestr, name=tzstr
        )

    if tzstr[0] == "-":
        return builder.build_timezone(
            negative=True, hh=hourstr, mm=minutestr, name=tzstr
        )

    raise ISOFormatError('"{0}" is not a valid ISO 8601 time offset.'.format(tzstr))



# -*- coding: utf-8 -*-

# Copyright (c) 2025, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

from aniso8601.builders import DatetimeTuple, DateTuple, TupleBuilder
from aniso8601.builders.python import PythonTimeBuilder
from aniso8601.compat import is_string
from aniso8601.date import parse_date
from aniso8601.duration import parse_duration
from aniso8601.exceptions import ISOFormatError
from aniso8601.resolution import IntervalResolution
from aniso8601.time import parse_datetime, parse_time


def get_interval_resolution(
    isointervalstr, intervaldelimiter="/", datetimedelimiter="T"
):
    isointervaltuple = parse_interval(
        isointervalstr,
        intervaldelimiter=intervaldelimiter,
        datetimedelimiter=datetimedelimiter,
        builder=TupleBuilder,
    )

    return _get_interval_resolution(isointervaltuple)


def get_repeating_interval_resolution(
    isointervalstr, intervaldelimiter="/", datetimedelimiter="T"
):
    repeatingintervaltuple = parse_repeating_interval(
        isointervalstr,
        intervaldelimiter=intervaldelimiter,
        datetimedelimiter=datetimedelimiter,
        builder=TupleBuilder,
    )

    return _get_interval_resolution(repeatingintervaltuple.interval)


def _get_interval_resolution(intervaltuple):
    if intervaltuple.start is not None and intervaltuple.end is not None:
        return max(
            _get_interval_component_resolution(intervaltuple.start),
            _get_interval_component_resolution(intervaltuple.end),
        )

    if intervaltuple.start is not None and intervaltuple.duration is not None:
        return max(
            _get_interval_component_resolution(intervaltuple.start),
            _get_interval_component_resolution(intervaltuple.duration),
        )

    return max(
        _get_interval_component_resolution(intervaltuple.end),
        _get_interval_component_resolution(intervaltuple.duration),
    )


def _get_interval_component_resolution(componenttuple):
    if isinstance(componenttuple, DateTuple):
        if componenttuple.DDD is not None:
            # YYYY-DDD
            # YYYYDDD
            return IntervalResolution.Ordinal

        if componenttuple.D is not None:
            # YYYY-Www-D
            # YYYYWwwD
            return IntervalResolution.Weekday

        if componenttuple.Www is not None:
            # YYYY-Www
            # YYYYWww
            return IntervalResolution.Week

        if componenttuple.DD is not None:
            # YYYY-MM-DD
            # YYYYMMDD
            return IntervalResolution.Day

        if componenttuple.MM is not None:
            # YYYY-MM
            return IntervalResolution.Month

        # Y[YYY]
        return IntervalResolution.Year

    if isinstance(componenttuple, DatetimeTuple):
        # Datetime
        if componenttuple.time.ss is not None:
            return IntervalResolution.Seconds

        if componenttuple.time.mm is not None:
            return IntervalResolution.Minutes

        return IntervalResolution.Hours

    # Duration
    if componenttuple.TnS is not None:
        return IntervalResolution.Seconds

    if componenttuple.TnM is not None:
        return IntervalResolution.Minutes

    if componenttuple.TnH is not None:
        return IntervalResolution.Hours

    if componenttuple.PnD is not None:
        return IntervalResolution.Day

    if componenttuple.PnW is not None:
        return IntervalResolution.Week

    if componenttuple.PnM is not None:
        return IntervalResolution.Month

    return IntervalResolution.Year


def parse_interval(
    isointervalstr,
    intervaldelimiter="/",
    datetimedelimiter="T",
    builder=PythonTimeBuilder,
):
    # Given a string representing an ISO 8601 interval, return an
    # interval built by the given builder. Valid formats are:
    #
    # <start>/<end>
    # <start>/<duration>
    # <duration>/<end>
    #
    # The <start> and <end> values can represent dates, or datetimes,
    # not times.
    #
    # The format:
    #
    # <duration>
    #
    # Is expressly not supported as there is no way to provide the additional
    # required context.

    if is_string(isointervalstr) is False:
        raise ValueError("Interval must be string.")

    if len(isointervalstr) == 0:
        raise ISOFormatError("Interval string is empty.")

    if isointervalstr[0] == "R":
        raise ISOFormatError(
            "ISO 8601 repeating intervals must be parsed "
            "with parse_repeating_interval."
        )

    intervaldelimitercount = isointervalstr.count(intervaldelimiter)

    if intervaldelimitercount == 0:
        raise ISOFormatError(
            'Interval delimiter "{0}" is not in interval '
            'string "{1}".'.format(intervaldelimiter, isointervalstr)
        )

    if intervaldelimitercount > 1:
        raise ISOFormatError(
            "{0} is not a valid ISO 8601 interval".format(isointervalstr)
        )

    return _parse_interval(
        isointervalstr, builder, intervaldelimiter, datetimedelimiter
    )


def parse_repeating_interval(
    isointervalstr,
    intervaldelimiter="/",
    datetimedelimiter="T",
    builder=PythonTimeBuilder,
):
    # Given a string representing an ISO 8601 interval repeating, return an
    # interval built by the given builder. Valid formats are:
    #
    # Rnn/<interval>
    # R/<interval>

    if not isinstance(isointervalstr, str):
        raise ValueError("Interval must be string.")

    if len(isointervalstr) == 0:
        raise ISOFormatError("Repeating interval string is empty.")

    if isointervalstr[0] != "R":
        raise ISOFormatError("ISO 8601 repeating interval must start with an R.")

    if intervaldelimiter not in isointervalstr:
        raise ISOFormatError(
            'Interval delimiter "{0}" is not in interval '
            'string "{1}".'.format(intervaldelimiter, isointervalstr)
        )

    # Parse the number of iterations
    iterationpart, intervalpart = isointervalstr.split(intervaldelimiter, 1)

    if len(iterationpart) > 1:
        R = False
        Rnn = iterationpart[1:]
    else:
        R = True
        Rnn = None

    interval = _parse_interval(
        intervalpart, TupleBuilder, intervaldelimiter, datetimedelimiter
    )

    return builder.build_repeating_interval(R=R, Rnn=Rnn, interval=interval)


def _parse_interval(
    isointervalstr, builder, intervaldelimiter="/", datetimedelimiter="T"
):
    # Returns a tuple containing the start of the interval, the end of the
    # interval, and or the interval duration

    firstpart, secondpart = isointervalstr.split(intervaldelimiter)

    if len(firstpart) == 0 or len(secondpart) == 0:
        raise ISOFormatError(
            "{0} is not a valid ISO 8601 interval".format(isointervalstr)
        )

    if firstpart[0] == "P":
        # <duration>/<end>
        # Notice that these are not returned 'in order' (earlier to later), this
        # is to maintain consistency with parsing <start>/<end> durations, as
        # well as making repeating interval code cleaner. Users who desire
        # durations to be in order can use the 'sorted' operator.
        duration = parse_duration(firstpart, builder=TupleBuilder)

        # We need to figure out if <end> is a date, or a datetime
        if secondpart.find(datetimedelimiter) != -1:
            # <end> is a datetime
            endtuple = parse_datetime(
                secondpart, delimiter=datetimedelimiter, builder=TupleBuilder
            )
        else:
            endtuple = parse_date(secondpart, builder=TupleBuilder)

        return builder.build_interval(end=endtuple, duration=duration)

    if secondpart[0] == "P":
        # <start>/<duration>
        # We need to figure out if <start> is a date, or a datetime
        duration = parse_duration(secondpart, builder=TupleBuilder)

        if firstpart.find(datetimedelimiter) != -1:
            # <start> is a datetime
            starttuple = parse_datetime(
                firstpart, delimiter=datetimedelimiter, builder=TupleBuilder
            )
        else:
            # <start> must just be a date
            starttuple = parse_date(firstpart, builder=TupleBuilder)

        return builder.build_interval(start=starttuple, duration=duration)

    # <start>/<end>
    if firstpart.find(datetimedelimiter) != -1:
        # Both parts are datetimes
        starttuple = parse_datetime(
            firstpart, delimiter=datetimedelimiter, builder=TupleBuilder
        )
    else:
        starttuple = parse_date(firstpart, builder=TupleBuilder)

    endtuple = _parse_interval_end(secondpart, starttuple, datetimedelimiter)

    return builder.build_interval(start=starttuple, end=endtuple)


def _parse_interval_end(endstr, starttuple, datetimedelimiter):
    datestr = None
    timestr = None

    monthstr = None
    daystr = None

    concise = False

    if isinstance(starttuple, DateTuple):
        startdatetuple = starttuple
    else:
        # Start is a datetime
        startdatetuple = starttuple.date

    if datetimedelimiter in endstr:
        datestr, timestr = endstr.split(datetimedelimiter, 1)
    elif ":" in endstr:
        timestr = endstr
    else:
        datestr = endstr

    if timestr is not None:
        endtimetuple = parse_time(timestr, builder=TupleBuilder)

        # End is just a time
        if datestr is None:
            return endtimetuple

    # Handle backwards concise representation
    if datestr.count("-") == 1:
        monthstr, daystr = datestr.split("-")
        concise = True
    elif len(datestr) <= 2:
        daystr = datestr
        concise = True
    elif len(datestr) <= 4:
        monthstr = datestr[0:2]
        daystr = datestr[2:]
        concise = True

    if concise is True:
        concisedatestr = startdatetuple.YYYY

        # Separators required because concise elements may be missing digits
        if monthstr is not None:
            concisedatestr += "-" + monthstr
        else:
            concisedatestr += "-" + startdatetuple.MM

        concisedatestr += "-" + daystr

        enddatetuple = parse_date(concisedatestr, builder=TupleBuilder)

        # Clear unsupplied components
        if monthstr is None:
            enddatetuple = TupleBuilder.build_date(DD=enddatetuple.DD)
        else:
            # Year not provided
            enddatetuple = TupleBuilder.build_date(
                MM=enddatetuple.MM, DD=enddatetuple.DD
            )
    else:
        enddatetuple = parse_date(datestr, builder=TupleBuilder)

    if timestr is None:
        return enddatetuple

    return TupleBuilder.build_datetime(enddatetuple, endtimetuple)



# -*- coding: utf-8 -*-

# Copyright (c) 2025, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

from aniso8601 import compat
from aniso8601.builders import TupleBuilder
from aniso8601.builders.python import PythonTimeBuilder
from aniso8601.date import parse_date
from aniso8601.decimalfraction import normalize
from aniso8601.exceptions import ISOFormatError
from aniso8601.resolution import DurationResolution
from aniso8601.time import parse_time


def get_duration_resolution(isodurationstr):
    # Valid string formats are:
    #
    # PnYnMnDTnHnMnS (or any reduced precision equivalent)
    # PnW
    # P<date>T<time>
    isodurationtuple = parse_duration(isodurationstr, builder=TupleBuilder)

    if isodurationtuple.TnS is not None:
        return DurationResolution.Seconds

    if isodurationtuple.TnM is not None:
        return DurationResolution.Minutes

    if isodurationtuple.TnH is not None:
        return DurationResolution.Hours

    if isodurationtuple.PnD is not None:
        return DurationResolution.Days

    if isodurationtuple.PnW is not None:
        return DurationResolution.Weeks

    if isodurationtuple.PnM is not None:
        return DurationResolution.Months

    return DurationResolution.Years


def parse_duration(isodurationstr, builder=PythonTimeBuilder):
    # Given a string representing an ISO 8601 duration, return a
    # a duration built by the given builder. Valid formats are:
    #
    # PnYnMnDTnHnMnS (or any reduced precision equivalent)
    # PnW
    # P<date>T<time>

    if compat.is_string(isodurationstr) is False:
        raise ValueError("Duration must be string.")

    if len(isodurationstr) == 0:
        raise ISOFormatError(
            '"{0}" is not a valid ISO 8601 duration.'.format(isodurationstr)
        )

    if isodurationstr[0] != "P":
        raise ISOFormatError("ISO 8601 duration must start with a P.")

    # If Y, M, D, H, S, or W are in the string,
    # assume it is a specified duration
    if _has_any_component(isodurationstr, ["Y", "M", "D", "H", "S", "W"]) is True:
        parseresult = _parse_duration_prescribed(isodurationstr)
        return builder.build_duration(**parseresult)

    if isodurationstr.find("T") != -1:
        parseresult = _parse_duration_combined(isodurationstr)
        return builder.build_duration(**parseresult)

    raise ISOFormatError(
        '"{0}" is not a valid ISO 8601 duration.'.format(isodurationstr)
    )


def _parse_duration_prescribed(isodurationstr):
    # durationstr can be of the form PnYnMnDTnHnMnS or PnW

    # Make sure the end character is valid
    # https://bitbucket.org/nielsenb/aniso8601/issues/9/durations-with-trailing-garbage-are-parsed
    if isodurationstr[-1] not in ["Y", "M", "D", "H", "S", "W"]:
        raise ISOFormatError("ISO 8601 duration must end with a valid character.")

    # Make sure only the lowest order element has decimal precision
    durationstr = normalize(isodurationstr)

    if durationstr.count(".") > 1:
        raise ISOFormatError(
            "ISO 8601 allows only lowest order element to have a decimal fraction."
        )

    seperatoridx = durationstr.find(".")

    if seperatoridx != -1:
        remaining = durationstr[seperatoridx + 1 : -1]

        # There should only ever be 1 letter after a decimal if there is more
        # then one, the string is invalid
        if remaining.isdigit() is False:
            raise ISOFormatError(
                "ISO 8601 duration must end with a single valid character."
            )

    # Do not allow W in combination with other designators
    # https://bitbucket.org/nielsenb/aniso8601/issues/2/week-designators-should-not-be-combinable
    if (
        durationstr.find("W") != -1
        and _has_any_component(durationstr, ["Y", "M", "D", "H", "S"]) is True
    ):
        raise ISOFormatError(
            "ISO 8601 week designators may not be combined "
            "with other time designators."
        )

    # Parse the elements of the duration
    if durationstr.find("T") == -1:
        return _parse_duration_prescribed_notime(durationstr)

    return _parse_duration_prescribed_time(durationstr)


def _parse_duration_prescribed_notime(isodurationstr):
    # durationstr can be of the form PnYnMnD or PnW

    durationstr = normalize(isodurationstr)

    yearstr = None
    monthstr = None
    daystr = None
    weekstr = None

    weekidx = durationstr.find("W")
    yearidx = durationstr.find("Y")
    monthidx = durationstr.find("M")
    dayidx = durationstr.find("D")

    if weekidx != -1:
        weekstr = durationstr[1:-1]
    elif yearidx != -1 and monthidx != -1 and dayidx != -1:
        yearstr = durationstr[1:yearidx]
        monthstr = durationstr[yearidx + 1 : monthidx]
        daystr = durationstr[monthidx + 1 : -1]
    elif yearidx != -1 and monthidx != -1:
        yearstr = durationstr[1:yearidx]
        monthstr = durationstr[yearidx + 1 : monthidx]
    elif yearidx != -1 and dayidx != -1:
        yearstr = durationstr[1:yearidx]
        daystr = durationstr[yearidx + 1 : dayidx]
    elif monthidx != -1 and dayidx != -1:
        monthstr = durationstr[1:monthidx]
        daystr = durationstr[monthidx + 1 : -1]
    elif yearidx != -1:
        yearstr = durationstr[1:-1]
    elif monthidx != -1:
        monthstr = durationstr[1:-1]
    elif dayidx != -1:
        daystr = durationstr[1:-1]
    else:
        raise ISOFormatError(
            '"{0}" is not a valid ISO 8601 duration.'.format(isodurationstr)
        )

    for componentstr in [yearstr, monthstr, daystr, weekstr]:
        if componentstr is not None:
            if "." in componentstr:
                intstr = componentstr.split(".", 1)[0]

                if intstr.isdigit() is False:
                    raise ISOFormatError(
                        '"{0}" is not a valid ISO 8601 duration.'.format(isodurationstr)
                    )
            else:
                if componentstr.isdigit() is False:
                    raise ISOFormatError(
                        '"{0}" is not a valid ISO 8601 duration.'.format(isodurationstr)
                    )

    return {"PnY": yearstr, "PnM": monthstr, "PnW": weekstr, "PnD": daystr}


def _parse_duration_prescribed_time(isodurationstr):
    # durationstr can be of the form PnYnMnDTnHnMnS

    timeidx = isodurationstr.find("T")

    datestr = isodurationstr[:timeidx]
    timestr = normalize(isodurationstr[timeidx + 1 :])

    hourstr = None
    minutestr = None
    secondstr = None

    houridx = timestr.find("H")
    minuteidx = timestr.find("M")
    secondidx = timestr.find("S")

    if houridx != -1 and minuteidx != -1 and secondidx != -1:
        hourstr = timestr[0:houridx]
        minutestr = timestr[houridx + 1 : minuteidx]
        secondstr = timestr[minuteidx + 1 : -1]
    elif houridx != -1 and minuteidx != -1:
        hourstr = timestr[0:houridx]
        minutestr = timestr[houridx + 1 : minuteidx]
    elif houridx != -1 and secondidx != -1:
        hourstr = timestr[0:houridx]
        secondstr = timestr[houridx + 1 : -1]
    elif minuteidx != -1 and secondidx != -1:
        minutestr = timestr[0:minuteidx]
        secondstr = timestr[minuteidx + 1 : -1]
    elif houridx != -1:
        hourstr = timestr[0:-1]
    elif minuteidx != -1:
        minutestr = timestr[0:-1]
    elif secondidx != -1:
        secondstr = timestr[0:-1]
    else:
        raise ISOFormatError(
            '"{0}" is not a valid ISO 8601 duration.'.format(isodurationstr)
        )

    for componentstr in [hourstr, minutestr, secondstr]:
        if componentstr is not None:
            if "." in componentstr:
                intstr = componentstr.split(".", 1)[0]

                if intstr.isdigit() is False:
                    raise ISOFormatError(
                        '"{0}" is not a valid ISO 8601 duration.'.format(isodurationstr)
                    )
            else:
                if componentstr.isdigit() is False:
                    raise ISOFormatError(
                        '"{0}" is not a valid ISO 8601 duration.'.format(isodurationstr)
                    )

    # Parse any date components
    durationdict = {"PnY": None, "PnM": None, "PnW": None, "PnD": None}

    if len(datestr) > 1:
        durationdict = _parse_duration_prescribed_notime(datestr)

    durationdict.update({"TnH": hourstr, "TnM": minutestr, "TnS": secondstr})

    return durationdict


def _parse_duration_combined(durationstr):
    # Period of the form P<date>T<time>

    # Split the string in to its component parts
    datepart, timepart = durationstr[1:].split("T", 1)  # We skip the 'P'

    datevalue = parse_date(datepart, builder=TupleBuilder)
    timevalue = parse_time(timepart, builder=TupleBuilder)

    return {
        "PnY": datevalue.YYYY,
        "PnM": datevalue.MM,
        "PnD": datevalue.DD,
        "TnH": timevalue.hh,
        "TnM": timevalue.mm,
        "TnS": timevalue.ss,
    }


def _has_any_component(durationstr, components):
    # Given a duration string, and a list of components, returns True
    # if any of the listed components are present, False otherwise.
    #
    # For instance:
    # durationstr = 'P1Y'
    # components = ['Y', 'M']
    #
    # returns True
    #
    # durationstr = 'P1Y'
    # components = ['M', 'D']
    #
    # returns False

    for component in components:
        if durationstr.find(component) != -1:
            return True

    return False



# -*- coding: utf-8 -*-

# Copyright (c) 2025, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import unittest

import aniso8601
from aniso8601.date import get_date_resolution, parse_date
from aniso8601.exceptions import DayOutOfBoundsError, ISOFormatError
from aniso8601.resolution import DateResolution
from aniso8601.tests.compat import mock


class TestDateResolutionFunctions(unittest.TestCase):
    def test_get_date_resolution_year(self):
        self.assertEqual(get_date_resolution("2013"), DateResolution.Year)
        self.assertEqual(get_date_resolution("0001"), DateResolution.Year)
        self.assertEqual(get_date_resolution("19"), DateResolution.Year)

    def test_get_date_resolution_month(self):
        self.assertEqual(get_date_resolution("1981-04"), DateResolution.Month)

    def test_get_date_resolution_week(self):
        self.assertEqual(get_date_resolution("2004-W53"), DateResolution.Week)
        self.assertEqual(get_date_resolution("2009-W01"), DateResolution.Week)
        self.assertEqual(get_date_resolution("2004W53"), DateResolution.Week)

    def test_get_date_resolution_day(self):
        self.assertEqual(get_date_resolution("2004-04-11"), DateResolution.Day)
        self.assertEqual(get_date_resolution("20090121"), DateResolution.Day)

    def test_get_date_resolution_year_weekday(self):
        self.assertEqual(get_date_resolution("2004-W53-6"), DateResolution.Weekday)
        self.assertEqual(get_date_resolution("2004W536"), DateResolution.Weekday)

    def test_get_date_resolution_year_ordinal(self):
        self.assertEqual(get_date_resolution("1981-095"), DateResolution.Ordinal)
        self.assertEqual(get_date_resolution("1981095"), DateResolution.Ordinal)

    def test_get_date_resolution_badtype(self):
        testtuples = (None, 1, False, 1.234)

        for testtuple in testtuples:
            with self.assertRaises(ValueError):
                get_date_resolution(testtuple)

    def test_get_date_resolution_extended_year(self):
        testtuples = ("+2000", "+30000")

        for testtuple in testtuples:
            with self.assertRaises(NotImplementedError):
                get_date_resolution(testtuple)

    def test_get_date_resolution_badweek(self):
        testtuples = ("2004-W1", "2004W1")

        for testtuple in testtuples:
            with self.assertRaises(ISOFormatError):
                get_date_resolution(testtuple)

    def test_get_date_resolution_badweekday(self):
        testtuples = ("2004-W53-67", "2004W5367")

        for testtuple in testtuples:
            with self.assertRaises(ISOFormatError):
                get_date_resolution(testtuple)

    def test_get_date_resolution_badstr(self):
        testtuples = (
            "W53",
            "2004-W",
            "2014-01-230",
            "2014-012-23",
            "201-01-23",
            "201401230",
            "201401",
            "",
        )

        for testtuple in testtuples:
            with self.assertRaises(ISOFormatError):
                get_date_resolution(testtuple)


class TestDateParserFunctions(unittest.TestCase):
    def test_parse_date(self):
        testtuples = (
            (
                "2013",
                {
                    "YYYY": "2013",
                    "MM": None,
                    "DD": None,
                    "Www": None,
                    "D": None,
                    "DDD": None,
                },
            ),
            (
                "0001",
                {
                    "YYYY": "0001",
                    "MM": None,
                    "DD": None,
                    "Www": None,
                    "D": None,
                    "DDD": None,
                },
            ),
            (
                "19",
                {
                    "YYYY": "19",
                    "MM": None,
                    "DD": None,
                    "Www": None,
                    "D": None,
                    "DDD": None,
                },
            ),
            (
                "10",
                {
                    "YYYY": "10",
                    "MM": None,
                    "DD": None,
                    "Www": None,
                    "D": None,
                    "DDD": None,
                },
            ),
            (
                "1981-04-05",
                {
                    "YYYY": "1981",
                    "MM": "04",
                    "DD": "05",
                    "Www": None,
                    "D": None,
                    "DDD": None,
                },
            ),
            (
                "19810405",
                {
                    "YYYY": "1981",
                    "MM": "04",
                    "DD": "05",
                    "Www": None,
                    "D": None,
                    "DDD": None,
                },
            ),
            (
                "1981-04",
                {
                    "YYYY": "1981",
                    "MM": "04",
                    "DD": None,
                    "Www": None,
                    "D": None,
                    "DDD": None,
                },
            ),
            (
                "2004-W53",
                {
                    "YYYY": "2004",
                    "MM": None,
                    "DD": None,
                    "Www": "53",
                    "D": None,
                    "DDD": None,
                },
            ),
            (
                "2009-W01",
                {
                    "YYYY": "2009",
                    "MM": None,
                    "DD": None,
                    "Www": "01",
                    "D": None,
                    "DDD": None,
                },
            ),
            (
                "2004-W53-6",
                {
                    "YYYY": "2004",
                    "MM": None,
                    "DD": None,
                    "Www": "53",
                    "D": "6",
                    "DDD": None,
                },
            ),
            (
                "2004W53",
                {
                    "YYYY": "2004",
                    "MM": None,
                    "DD": None,
                    "Www": "53",
                    "D": None,
                    "DDD": None,
                },
            ),
            (
                "2004W536",
                {
                    "YYYY": "2004",
                    "MM": None,
                    "DD": None,
                    "Www": "53",
                    "D": "6",
                    "DDD": None,
                },
            ),
            (
                "1981-095",
                {
                    "YYYY": "1981",
                    "MM": None,
                    "DD": None,
                    "Www": None,
                    "D": None,
                    "DDD": "095",
                },
            ),
            (
                "1981095",
                {
                    "YYYY": "1981",
                    "MM": None,
                    "DD": None,
                    "Www": None,
                    "D": None,
                    "DDD": "095",
                },
            ),
            (
                "1980366",
                {
                    "YYYY": "1980",
                    "MM": None,
                    "DD": None,
                    "Www": None,
                    "D": None,
                    "DDD": "366",
                },
            ),
        )

        for testtuple in testtuples:
            with mock.patch.object(
                aniso8601.date.PythonTimeBuilder, "build_date"
            ) as mockBuildDate:
                mockBuildDate.return_value = testtuple[1]

                result = parse_date(testtuple[0])

                self.assertEqual(result, testtuple[1])
                mockBuildDate.assert_called_once_with(**testtuple[1])

    def test_parse_date_badtype(self):
        testtuples = (None, 1, False, 1.234)

        for testtuple in testtuples:
            with self.assertRaises(ValueError):
                parse_date(testtuple, builder=None)

    def test_parse_date_badstr(self):
        testtuples = (
            "1",
            "123" "W53",
            "2004-W",
            "2014-01-230",
            "2014-012-23",
            "201-01-23",
            "201401230",
            "201401",
            "9999 W53",
            "20.50230",
            "198104",
            "bad",
            "",
        )

        for testtuple in testtuples:
            with self.assertRaises(ISOFormatError):
                parse_date(testtuple, builder=None)

    def test_parse_date_mockbuilder(self):
        mockBuilder = mock.Mock()

        expectedargs = {
            "YYYY": "1981",
            "MM": "04",
            "DD": "05",
            "Www": None,
            "D": None,
            "DDD": None,
        }

        mockBuilder.build_date.return_value = expectedargs

        result = parse_date("1981-04-05", builder=mockBuilder)

        self.assertEqual(result, expectedargs)
        mockBuilder.build_date.assert_called_once_with(**expectedargs)



# -*- coding: utf-8 -*-

# Copyright (c) 2025, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import datetime
from collections import namedtuple
from functools import partial

from aniso8601.builders import (
    BaseTimeBuilder,
    DateTuple,
    Limit,
    TupleBuilder,
    cast,
    range_check,
)
from aniso8601.exceptions import (
    DayOutOfBoundsError,
    HoursOutOfBoundsError,
    MinutesOutOfBoundsError,
    MonthOutOfBoundsError,
    SecondsOutOfBoundsError,
    WeekOutOfBoundsError,
    YearOutOfBoundsError,
)
from aniso8601.utcoffset import UTCOffset

DAYS_PER_YEAR = 365
DAYS_PER_MONTH = 30
DAYS_PER_WEEK = 7

HOURS_PER_DAY = 24

MINUTES_PER_HOUR = 60
MINUTES_PER_DAY = MINUTES_PER_HOUR * HOURS_PER_DAY

SECONDS_PER_MINUTE = 60
SECONDS_PER_DAY = MINUTES_PER_DAY * SECONDS_PER_MINUTE

MICROSECONDS_PER_SECOND = int(1e6)

MICROSECONDS_PER_MINUTE = 60 * MICROSECONDS_PER_SECOND
MICROSECONDS_PER_HOUR = 60 * MICROSECONDS_PER_MINUTE
MICROSECONDS_PER_DAY = 24 * MICROSECONDS_PER_HOUR
MICROSECONDS_PER_WEEK = 7 * MICROSECONDS_PER_DAY
MICROSECONDS_PER_MONTH = DAYS_PER_MONTH * MICROSECONDS_PER_DAY
MICROSECONDS_PER_YEAR = DAYS_PER_YEAR * MICROSECONDS_PER_DAY

TIMEDELTA_MAX_DAYS = datetime.timedelta.max.days

FractionalComponent = namedtuple(
    "FractionalComponent", ["principal", "microsecondremainder"]
)


def year_range_check(valuestr, limit):
    YYYYstr = valuestr

    # Truncated dates, like '19', refer to 1900-1999 inclusive,
    # we simply parse to 1900, Y and YYY strings are not supported
    if len(valuestr) == 2:
        # Shift 0s in from the left to form complete year
        YYYYstr = valuestr.ljust(4, "0")

    return range_check(YYYYstr, limit)


def fractional_range_check(conversion, valuestr, limit):
    if valuestr is None:
        return None

    if "." in valuestr:
        castfunc = partial(_cast_to_fractional_component, conversion)
    else:
        castfunc = int

    value = cast(valuestr, castfunc, thrownmessage=limit.casterrorstring)

    if isinstance(value, FractionalComponent):
        tocheck = float(valuestr)
    else:
        tocheck = int(valuestr)

    if limit.min is not None and tocheck < limit.min:
        raise limit.rangeexception(limit.rangeerrorstring)

    if limit.max is not None and tocheck > limit.max:
        raise limit.rangeexception(limit.rangeerrorstring)

    return value


def _cast_to_fractional_component(conversion, floatstr):
    # Splits a string with a decimal point into an int, and
    # int representing the floating point remainder as a number
    # of microseconds, determined by multiplying by conversion
    intpart, floatpart = floatstr.split(".")

    intvalue = int(intpart)
    preconvertedvalue = int(floatpart)

    convertedvalue = (preconvertedvalue * conversion) // (10 ** len(floatpart))

    return FractionalComponent(intvalue, convertedvalue)


class PythonTimeBuilder(BaseTimeBuilder):
    # 0000 (1 BC) is not representable as a Python date
    DATE_YYYY_LIMIT = Limit(
        "Invalid year string.",
        datetime.MINYEAR,
        datetime.MAXYEAR,
        YearOutOfBoundsError,
        "Year must be between {0}..{1}.".format(datetime.MINYEAR, datetime.MAXYEAR),
        year_range_check,
    )
    TIME_HH_LIMIT = Limit(
        "Invalid hour string.",
        0,
        24,
        HoursOutOfBoundsError,
        "Hour must be between 0..24 with 24 representing midnight.",
        partial(fractional_range_check, MICROSECONDS_PER_HOUR),
    )
    TIME_MM_LIMIT = Limit(
        "Invalid minute string.",
        0,
        59,
        MinutesOutOfBoundsError,
        "Minute must be between 0..59.",
        partial(fractional_range_check, MICROSECONDS_PER_MINUTE),
    )
    TIME_SS_LIMIT = Limit(
        "Invalid second string.",
        0,
        60,
        SecondsOutOfBoundsError,
        "Second must be between 0..60 with 60 representing a leap second.",
        partial(fractional_range_check, MICROSECONDS_PER_SECOND),
    )
    DURATION_PNY_LIMIT = Limit(
        "Invalid year duration string.",
        None,
        None,
        YearOutOfBoundsError,
        None,
        partial(fractional_range_check, MICROSECONDS_PER_YEAR),
    )
    DURATION_PNM_LIMIT = Limit(
        "Invalid month duration string.",
        None,
        None,
        MonthOutOfBoundsError,
        None,
        partial(fractional_range_check, MICROSECONDS_PER_MONTH),
    )
    DURATION_PNW_LIMIT = Limit(
        "Invalid week duration string.",
        None,
        None,
        WeekOutOfBoundsError,
        None,
        partial(fractional_range_check, MICROSECONDS_PER_WEEK),
    )
    DURATION_PND_LIMIT = Limit(
        "Invalid day duration string.",
        None,
        None,
        DayOutOfBoundsError,
        None,
        partial(fractional_range_check, MICROSECONDS_PER_DAY),
    )
    DURATION_TNH_LIMIT = Limit(
        "Invalid hour duration string.",
        None,
        None,
        HoursOutOfBoundsError,
        None,
        partial(fractional_range_check, MICROSECONDS_PER_HOUR),
    )
    DURATION_TNM_LIMIT = Limit(
        "Invalid minute duration string.",
        None,
        None,
        MinutesOutOfBoundsError,
        None,
        partial(fractional_range_check, MICROSECONDS_PER_MINUTE),
    )
    DURATION_TNS_LIMIT = Limit(
        "Invalid second duration string.",
        None,
        None,
        SecondsOutOfBoundsError,
        None,
        partial(fractional_range_check, MICROSECONDS_PER_SECOND),
    )

    DATE_RANGE_DICT = BaseTimeBuilder.DATE_RANGE_DICT
    DATE_RANGE_DICT["YYYY"] = DATE_YYYY_LIMIT

    TIME_RANGE_DICT = {"hh": TIME_HH_LIMIT, "mm": TIME_MM_LIMIT, "ss": TIME_SS_LIMIT}

    DURATION_RANGE_DICT = {
        "PnY": DURATION_PNY_LIMIT,
        "PnM": DURATION_PNM_LIMIT,
        "PnW": DURATION_PNW_LIMIT,
        "PnD": DURATION_PND_LIMIT,
        "TnH": DURATION_TNH_LIMIT,
        "TnM": DURATION_TNM_LIMIT,
        "TnS": DURATION_TNS_LIMIT,
    }

    @classmethod
    def build_date(cls, YYYY=None, MM=None, DD=None, Www=None, D=None, DDD=None):
        YYYY, MM, DD, Www, D, DDD = cls.range_check_date(YYYY, MM, DD, Www, D, DDD)

        if MM is None:
            MM = 1

        if DD is None:
            DD = 1

        if DDD is not None:
            return PythonTimeBuilder._build_ordinal_date(YYYY, DDD)

        if Www is not None:
            return PythonTimeBuilder._build_week_date(YYYY, Www, isoday=D)

        return datetime.date(YYYY, MM, DD)

    @classmethod
    def build_time(cls, hh=None, mm=None, ss=None, tz=None):
        # Builds a time from the given parts, handling fractional arguments
        # where necessary
        hours = 0
        minutes = 0
        seconds = 0
        microseconds = 0

        hh, mm, ss, tz = cls.range_check_time(hh, mm, ss, tz)

        if isinstance(hh, FractionalComponent):
            hours = hh.principal
            microseconds = hh.microsecondremainder
        elif hh is not None:
            hours = hh

        if isinstance(mm, FractionalComponent):
            minutes = mm.principal
            microseconds = mm.microsecondremainder
        elif mm is not None:
            minutes = mm

        if isinstance(ss, FractionalComponent):
            seconds = ss.principal
            microseconds = ss.microsecondremainder
        elif ss is not None:
            seconds = ss

        (
            hours,
            minutes,
            seconds,
            microseconds,
        ) = PythonTimeBuilder._distribute_microseconds(
            microseconds,
            (hours, minutes, seconds),
            (MICROSECONDS_PER_HOUR, MICROSECONDS_PER_MINUTE, MICROSECONDS_PER_SECOND),
        )

        # Move midnight into range
        if hours == 24:
            hours = 0

        # Datetimes don't handle fractional components, so we use a timedelta
        if tz is not None:
            return (
                datetime.datetime(
                    1, 1, 1, hour=hours, minute=minutes, tzinfo=cls._build_object(tz)
                )
                + datetime.timedelta(seconds=seconds, microseconds=microseconds)
            ).timetz()

        return (
            datetime.datetime(1, 1, 1, hour=hours, minute=minutes)
            + datetime.timedelta(seconds=seconds, microseconds=microseconds)
        ).time()

    @classmethod
    def build_datetime(cls, date, time):
        return datetime.datetime.combine(
            cls._build_object(date), cls._build_object(time)
        )

    @classmethod
    def build_duration(
        cls, PnY=None, PnM=None, PnW=None, PnD=None, TnH=None, TnM=None, TnS=None
    ):
        # PnY and PnM will be distributed to PnD, microsecond remainder to TnS
        PnY, PnM, PnW, PnD, TnH, TnM, TnS = cls.range_check_duration(
            PnY, PnM, PnW, PnD, TnH, TnM, TnS
        )

        seconds = TnS.principal
        microseconds = TnS.microsecondremainder

        return datetime.timedelta(
            days=PnD,
            seconds=seconds,
            microseconds=microseconds,
            minutes=TnM,
            hours=TnH,
            weeks=PnW,
        )

    @classmethod
    def build_interval(cls, start=None, end=None, duration=None):
        start, end, duration = cls.range_check_interval(start, end, duration)

        if start is not None and end is not None:
            # <start>/<end>
            startobject = cls._build_object(start)
            endobject = cls._build_object(end)

            return (startobject, endobject)

        durationobject = cls._build_object(duration)

        # Determine if datetime promotion is required
        datetimerequired = (
            duration.TnH is not None
            or duration.TnM is not None
            or duration.TnS is not None
            or durationobject.seconds != 0
            or durationobject.microseconds != 0
        )

        if end is not None:
            # <duration>/<end>
            endobject = cls._build_object(end)

            # Range check
            if isinstance(end, DateTuple) and datetimerequired is True:
                # <end> is a date, and <duration> requires datetime resolution
                return (
                    endobject,
                    cls.build_datetime(end, TupleBuilder.build_time()) - durationobject,
                )

            return (endobject, endobject - durationobject)

        # <start>/<duration>
        startobject = cls._build_object(start)

        # Range check
        if isinstance(start, DateTuple) and datetimerequired is True:
            # <start> is a date, and <duration> requires datetime resolution
            return (
                startobject,
                cls.build_datetime(start, TupleBuilder.build_time()) + durationobject,
            )

        return (startobject, startobject + durationobject)

    @classmethod
    def build_repeating_interval(cls, R=None, Rnn=None, interval=None):
        startobject = None
        endobject = None

        R, Rnn, interval = cls.range_check_repeating_interval(R, Rnn, interval)

        if interval.start is not None:
            startobject = cls._build_object(interval.start)

        if interval.end is not None:
            endobject = cls._build_object(interval.end)

        if interval.duration is not None:
            durationobject = cls._build_object(interval.duration)
        else:
            durationobject = endobject - startobject

        if R is True:
            if startobject is not None:
                return cls._date_generator_unbounded(startobject, durationobject)

            return cls._date_generator_unbounded(endobject, -durationobject)

        iterations = int(Rnn)

        if startobject is not None:
            return cls._date_generator(startobject, durationobject, iterations)

        return cls._date_generator(endobject, -durationobject, iterations)

    @classmethod
    def build_timezone(cls, negative=None, Z=None, hh=None, mm=None, name=""):
        negative, Z, hh, mm, name = cls.range_check_timezone(negative, Z, hh, mm, name)

        if Z is True:
            # Z -> UTC
            return UTCOffset(name="UTC", minutes=0)

        tzhour = int(hh)

        if mm is not None:
            tzminute = int(mm)
        else:
            tzminute = 0

        if negative is True:
            return UTCOffset(name=name, minutes=-(tzhour * 60 + tzminute))

        return UTCOffset(name=name, minutes=tzhour * 60 + tzminute)

    @classmethod
    def range_check_duration(
        cls,
        PnY=None,
        PnM=None,
        PnW=None,
        PnD=None,
        TnH=None,
        TnM=None,
        TnS=None,
        rangedict=None,
    ):
        years = 0
        months = 0
        days = 0
        weeks = 0
        hours = 0
        minutes = 0
        seconds = 0
        microseconds = 0

        PnY, PnM, PnW, PnD, TnH, TnM, TnS = BaseTimeBuilder.range_check_duration(
            PnY, PnM, PnW, PnD, TnH, TnM, TnS, rangedict=cls.DURATION_RANGE_DICT
        )

        if PnY is not None:
            if isinstance(PnY, FractionalComponent):
                years = PnY.principal
                microseconds = PnY.microsecondremainder
            else:
                years = PnY

            if years * DAYS_PER_YEAR > TIMEDELTA_MAX_DAYS:
                raise YearOutOfBoundsError("Duration exceeds maximum timedelta size.")

        if PnM is not None:
            if isinstance(PnM, FractionalComponent):
                months = PnM.principal
                microseconds = PnM.microsecondremainder
            else:
                months = PnM

            if months * DAYS_PER_MONTH > TIMEDELTA_MAX_DAYS:
                raise MonthOutOfBoundsError("Duration exceeds maximum timedelta size.")

        if PnW is not None:
            if isinstance(PnW, FractionalComponent):
                weeks = PnW.principal
                microseconds = PnW.microsecondremainder
            else:
                weeks = PnW

            if weeks * DAYS_PER_WEEK > TIMEDELTA_MAX_DAYS:
                raise WeekOutOfBoundsError("Duration exceeds maximum timedelta size.")

        if PnD is not None:
            if isinstance(PnD, FractionalComponent):
                days = PnD.principal
                microseconds = PnD.microsecondremainder
            else:
                days = PnD

            if days > TIMEDELTA_MAX_DAYS:
                raise DayOutOfBoundsError("Duration exceeds maximum timedelta size.")

        if TnH is not None:
            if isinstance(TnH, FractionalComponent):
                hours = TnH.principal
                microseconds = TnH.microsecondremainder
            else:
                hours = TnH

            if hours // HOURS_PER_DAY > TIMEDELTA_MAX_DAYS:
                raise HoursOutOfBoundsError("Duration exceeds maximum timedelta size.")

        if TnM is not None:
            if isinstance(TnM, FractionalComponent):
                minutes = TnM.principal
                microseconds = TnM.microsecondremainder
            else:
                minutes = TnM

            if minutes // MINUTES_PER_DAY > TIMEDELTA_MAX_DAYS:
                raise MinutesOutOfBoundsError(
                    "Duration exceeds maximum timedelta size."
                )

        if TnS is not None:
            if isinstance(TnS, FractionalComponent):
                seconds = TnS.principal
                microseconds = TnS.microsecondremainder
            else:
                seconds = TnS

            if seconds // SECONDS_PER_DAY > TIMEDELTA_MAX_DAYS:
                raise SecondsOutOfBoundsError(
                    "Duration exceeds maximum timedelta size."
                )

        (
            years,
            months,
            weeks,
            days,
            hours,
            minutes,
            seconds,
            microseconds,
        ) = PythonTimeBuilder._distribute_microseconds(
            microseconds,
            (years, months, weeks, days, hours, minutes, seconds),
            (
                MICROSECONDS_PER_YEAR,
                MICROSECONDS_PER_MONTH,
                MICROSECONDS_PER_WEEK,
                MICROSECONDS_PER_DAY,
                MICROSECONDS_PER_HOUR,
                MICROSECONDS_PER_MINUTE,
                MICROSECONDS_PER_SECOND,
            ),
        )

        # Note that weeks can be handled without conversion to days
        totaldays = years * DAYS_PER_YEAR + months * DAYS_PER_MONTH + days

        # Check against timedelta limits
        if (
            totaldays
            + weeks * DAYS_PER_WEEK
            + hours // HOURS_PER_DAY
            + minutes // MINUTES_PER_DAY
            + seconds // SECONDS_PER_DAY
            > TIMEDELTA_MAX_DAYS
        ):
            raise DayOutOfBoundsError("Duration exceeds maximum timedelta size.")

        return (
            None,
            None,
            weeks,
            totaldays,
            hours,
            minutes,
            FractionalComponent(seconds, microseconds),
        )

    @classmethod
    def range_check_interval(cls, start=None, end=None, duration=None):
        # Handles concise format, range checks any potential durations
        if start is not None and end is not None:
            # <start>/<end>
            # Handle concise format
            if cls._is_interval_end_concise(end) is True:
                end = cls._combine_concise_interval_tuples(start, end)

            return (start, end, duration)

        durationobject = cls._build_object(duration)

        if end is not None:
            # <duration>/<end>
            endobject = cls._build_object(end)

            # Range check
            if isinstance(end, DateTuple):
                enddatetime = cls.build_datetime(end, TupleBuilder.build_time())

                if enddatetime - datetime.datetime.min < durationobject:
                    raise YearOutOfBoundsError("Interval end less than minimium date.")
            else:
                mindatetime = datetime.datetime.min

                if end.time.tz is not None:
                    mindatetime = mindatetime.replace(tzinfo=endobject.tzinfo)

                if endobject - mindatetime < durationobject:
                    raise YearOutOfBoundsError("Interval end less than minimium date.")
        else:
            # <start>/<duration>
            startobject = cls._build_object(start)

            # Range check
            if type(start) is DateTuple:
                startdatetime = cls.build_datetime(start, TupleBuilder.build_time())

                if datetime.datetime.max - startdatetime < durationobject:
                    raise YearOutOfBoundsError(
                        "Interval end greater than maximum date."
                    )
            else:
                maxdatetime = datetime.datetime.max

                if start.time.tz is not None:
                    maxdatetime = maxdatetime.replace(tzinfo=startobject.tzinfo)

                if maxdatetime - startobject < durationobject:
                    raise YearOutOfBoundsError(
                        "Interval end greater than maximum date."
                    )

        return (start, end, duration)

    @staticmethod
    def _build_week_date(isoyear, isoweek, isoday=None):
        if isoday is None:
            return PythonTimeBuilder._iso_year_start(isoyear) + datetime.timedelta(
                weeks=isoweek - 1
            )

        return PythonTimeBuilder._iso_year_start(isoyear) + datetime.timedelta(
            weeks=isoweek - 1, days=isoday - 1
        )

    @staticmethod
    def _build_ordinal_date(isoyear, isoday):
        # Day of year to a date
        # https://stackoverflow.com/questions/2427555/python-question-year-and-day-of-year-to-date
        builtdate = datetime.date(isoyear, 1, 1) + datetime.timedelta(days=isoday - 1)

        return builtdate

    @staticmethod
    def _iso_year_start(isoyear):
        # Given an ISO year, returns the equivalent of the start of the year
        # on the Gregorian calendar (which is used by Python)
        # Stolen from:
        # http://stackoverflow.com/questions/304256/whats-the-best-way-to-find-the-inverse-of-datetime-isocalendar

        # Determine the location of the 4th of January, the first week of
        # the ISO year is the week containing the 4th of January
        # http://en.wikipedia.org/wiki/ISO_week_date
        fourth_jan = datetime.date(isoyear, 1, 4)

        # Note the conversion from ISO day (1 - 7) and Python day (0 - 6)
        delta = datetime.timedelta(days=fourth_jan.isoweekday() - 1)

        # Return the start of the year
        return fourth_jan - delta

    @staticmethod
    def _date_generator(startdate, timedelta, iterations):
        currentdate = startdate
        currentiteration = 0

        while currentiteration < iterations:
            yield currentdate

            # Update the values
            currentdate += timedelta
            currentiteration += 1

    @staticmethod
    def _date_generator_unbounded(startdate, timedelta):
        currentdate = startdate

        while True:
            yield currentdate

            # Update the value
            currentdate += timedelta

    @staticmethod
    def _distribute_microseconds(todistribute, recipients, reductions):
        # Given a number of microseconds as int, a tuple of ints length n
        # to distribute to, and a tuple of ints length n to divide todistribute
        # by (from largest to smallest), returns a tuple of length n + 1, with
        # todistribute divided across recipients using the reductions, with
        # the final remainder returned as the final tuple member
        results = []

        remainder = todistribute

        for index, reduction in enumerate(reductions):
            additional, remainder = divmod(remainder, reduction)

            results.append(recipients[index] + additional)

        # Always return the remaining microseconds
        results.append(remainder)

        return tuple(results)



# -*- coding: utf-8 -*-

# Copyright (c) 2025, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import unittest

import aniso8601
from aniso8601.builders import (
    BaseTimeBuilder,
    DatetimeTuple,
    DateTuple,
    DurationTuple,
    IntervalTuple,
    RepeatingIntervalTuple,
    TimeTuple,
    TimezoneTuple,
    TupleBuilder,
    cast,
)
from aniso8601.exceptions import (
    DayOutOfBoundsError,
    HoursOutOfBoundsError,
    ISOFormatError,
    LeapSecondError,
    MidnightBoundsError,
    MinutesOutOfBoundsError,
    MonthOutOfBoundsError,
    SecondsOutOfBoundsError,
    WeekOutOfBoundsError,
)
from aniso8601.tests.compat import mock


class LeapSecondSupportingTestBuilder(BaseTimeBuilder):
    LEAP_SECONDS_SUPPORTED = True


class TestBuilderFunctions(unittest.TestCase):
    def test_cast(self):
        self.assertEqual(cast("1", int), 1)
        self.assertEqual(cast("-2", int), -2)
        self.assertEqual(cast("3", float), float(3))
        self.assertEqual(cast("-4", float), float(-4))
        self.assertEqual(cast("5.6", float), 5.6)
        self.assertEqual(cast("-7.8", float), -7.8)

    def test_cast_exception(self):
        with self.assertRaises(ISOFormatError):
            cast("asdf", int)

        with self.assertRaises(ISOFormatError):
            cast("asdf", float)

    def test_cast_caughtexception(self):
        def tester(value):
            raise RuntimeError

        with self.assertRaises(ISOFormatError):
            cast("asdf", tester, caughtexceptions=(RuntimeError,))

    def test_cast_thrownexception(self):
        with self.assertRaises(RuntimeError):
            cast("asdf", int, thrownexception=RuntimeError)


class TestBaseTimeBuilder(unittest.TestCase):
    def test_build_date(self):
        with self.assertRaises(NotImplementedError):
            BaseTimeBuilder.build_date()

    def test_build_time(self):
        with self.assertRaises(NotImplementedError):
            BaseTimeBuilder.build_time()

    def test_build_datetime(self):
        with self.assertRaises(NotImplementedError):
            BaseTimeBuilder.build_datetime(None, None)

    def test_build_duration(self):
        with self.assertRaises(NotImplementedError):
            BaseTimeBuilder.build_duration()

    def test_build_interval(self):
        with self.assertRaises(NotImplementedError):
            BaseTimeBuilder.build_interval()

    def test_build_repeating_interval(self):
        with self.assertRaises(NotImplementedError):
            BaseTimeBuilder.build_repeating_interval()

    def test_build_timezone(self):
        with self.assertRaises(NotImplementedError):
            BaseTimeBuilder.build_timezone()

    def test_range_check_date(self):
        # Check the calendar for day ranges
        with self.assertRaises(DayOutOfBoundsError):
            BaseTimeBuilder.range_check_date(YYYY="0007", MM="02", DD="30")

        with self.assertRaises(DayOutOfBoundsError):
            BaseTimeBuilder.range_check_date(YYYY="0007", DDD="366")

        with self.assertRaises(MonthOutOfBoundsError):
            BaseTimeBuilder.range_check_date(YYYY="4333", MM="30", DD="30")

        # 0 isn't a valid week number
        with self.assertRaises(WeekOutOfBoundsError):
            BaseTimeBuilder.range_check_date(YYYY="2003", Www="00")

        # Week must not be larger than 53
        with self.assertRaises(WeekOutOfBoundsError):
            BaseTimeBuilder.range_check_date(YYYY="2004", Www="54")

        # 0 isn't a valid day number
        with self.assertRaises(DayOutOfBoundsError):
            BaseTimeBuilder.range_check_date(YYYY="2001", Www="02", D="0")

        # Day must not be larger than 7
        with self.assertRaises(DayOutOfBoundsError):
            BaseTimeBuilder.range_check_date(YYYY="2001", Www="02", D="8")

        with self.assertRaises(DayOutOfBoundsError):
            BaseTimeBuilder.range_check_date(YYYY="1981", DDD="000")

        # Day must be 365, or 366, not larger
        with self.assertRaises(DayOutOfBoundsError):
            BaseTimeBuilder.range_check_date(YYYY="1234", DDD="000")

        with self.assertRaises(DayOutOfBoundsError):
            BaseTimeBuilder.range_check_date(YYYY="1234", DDD="367")

        # https://bitbucket.org/nielsenb/aniso8601/issues/14/parsing-ordinal-dates-should-only-allow
        with self.assertRaises(DayOutOfBoundsError):
            BaseTimeBuilder.range_check_date(YYYY="1981", DDD="366")

        # Make sure Nones pass through unmodified
        self.assertEqual(
            BaseTimeBuilder.range_check_date(rangedict={}),
            (None, None, None, None, None, None),
        )

    def test_range_check_time(self):
        # Leap seconds not supported
        # https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
        # https://bitbucket.org/nielsenb/aniso8601/issues/13/parsing-of-leap-second-gives-wildly
        with self.assertRaises(LeapSecondError):
            BaseTimeBuilder.range_check_time(hh="23", mm="59", ss="60")

        with self.assertRaises(SecondsOutOfBoundsError):
            BaseTimeBuilder.range_check_time(hh="00", mm="00", ss="60")

        with self.assertRaises(SecondsOutOfBoundsError):
            BaseTimeBuilder.range_check_time(hh="00", mm="00", ss="61")

        with self.assertRaises(MinutesOutOfBoundsError):
            BaseTimeBuilder.range_check_time(hh="00", mm="61")

        with self.assertRaises(MinutesOutOfBoundsError):
            BaseTimeBuilder.range_check_time(hh="00", mm="60")

        with self.assertRaises(MinutesOutOfBoundsError):
            BaseTimeBuilder.range_check_time(hh="00", mm="60.1")

        with self.assertRaises(HoursOutOfBoundsError):
            BaseTimeBuilder.range_check_time(hh="25")

        # Hour 24 can only represent midnight
        with self.assertRaises(MidnightBoundsError):
            BaseTimeBuilder.range_check_time(hh="24", mm="00", ss="01")

        with self.assertRaises(MidnightBoundsError):
            BaseTimeBuilder.range_check_time(hh="24", mm="00.1")

        with self.assertRaises(MidnightBoundsError):
            BaseTimeBuilder.range_check_time(hh="24", mm="01")

        with self.assertRaises(MidnightBoundsError):
            BaseTimeBuilder.range_check_time(hh="24.1")

        # Leap seconds not supported
        # https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
        # https://bitbucket.org/nielsenb/aniso8601/issues/13/parsing-of-leap-second-gives-wildly
        with self.assertRaises(LeapSecondError):
            BaseTimeBuilder.range_check_time(hh="23", mm="59", ss="60")

        # Make sure Nones pass through unmodified
        self.assertEqual(
            BaseTimeBuilder.range_check_time(rangedict={}), (None, None, None, None)
        )

    def test_range_check_time_leap_seconds_supported(self):
        self.assertEqual(
            LeapSecondSupportingTestBuilder.range_check_time(hh="23", mm="59", ss="60"),
            (23, 59, 60, None),
        )

        with self.assertRaises(SecondsOutOfBoundsError):
            LeapSecondSupportingTestBuilder.range_check_time(hh="01", mm="02", ss="60")

    def test_range_check_duration(self):
        self.assertEqual(
            BaseTimeBuilder.range_check_duration(),
            (None, None, None, None, None, None, None),
        )

        self.assertEqual(
            BaseTimeBuilder.range_check_duration(rangedict={}),
            (None, None, None, None, None, None, None),
        )

    def test_range_check_repeating_interval(self):
        self.assertEqual(
            BaseTimeBuilder.range_check_repeating_interval(), (None, None, None)
        )

        self.assertEqual(
            BaseTimeBuilder.range_check_repeating_interval(rangedict={}),
            (None, None, None),
        )

    def test_range_check_timezone(self):
        self.assertEqual(
            BaseTimeBuilder.range_check_timezone(), (None, None, None, None, "")
        )

        self.assertEqual(
            BaseTimeBuilder.range_check_timezone(rangedict={}),
            (None, None, None, None, ""),
        )

    def test_build_object(self):
        datetest = (
            DateTuple("1", "2", "3", "4", "5", "6"),
            {"YYYY": "1", "MM": "2", "DD": "3", "Www": "4", "D": "5", "DDD": "6"},
        )

        timetest = (
            TimeTuple("1", "2", "3", TimezoneTuple(False, False, "4", "5", "tz name")),
            {
                "hh": "1",
                "mm": "2",
                "ss": "3",
                "tz": TimezoneTuple(False, False, "4", "5", "tz name"),
            },
        )

        datetimetest = (
            DatetimeTuple(
                DateTuple("1", "2", "3", "4", "5", "6"),
                TimeTuple(
                    "7", "8", "9", TimezoneTuple(True, False, "10", "11", "tz name")
                ),
            ),
            (
                DateTuple("1", "2", "3", "4", "5", "6"),
                TimeTuple(
                    "7", "8", "9", TimezoneTuple(True, False, "10", "11", "tz name")
                ),
            ),
        )

        durationtest = (
            DurationTuple("1", "2", "3", "4", "5", "6", "7"),
            {
                "PnY": "1",
                "PnM": "2",
                "PnW": "3",
                "PnD": "4",
                "TnH": "5",
                "TnM": "6",
                "TnS": "7",
            },
        )

        intervaltests = (
            (
                IntervalTuple(
                    DateTuple("1", "2", "3", "4", "5", "6"),
                    DateTuple("7", "8", "9", "10", "11", "12"),
                    None,
                ),
                {
                    "start": DateTuple("1", "2", "3", "4", "5", "6"),
                    "end": DateTuple("7", "8", "9", "10", "11", "12"),
                    "duration": None,
                },
            ),
            (
                IntervalTuple(
                    DateTuple("1", "2", "3", "4", "5", "6"),
                    None,
                    DurationTuple("7", "8", "9", "10", "11", "12", "13"),
                ),
                {
                    "start": DateTuple("1", "2", "3", "4", "5", "6"),
                    "end": None,
                    "duration": DurationTuple("7", "8", "9", "10", "11", "12", "13"),
                },
            ),
            (
                IntervalTuple(
                    None,
                    TimeTuple(
                        "1", "2", "3", TimezoneTuple(True, False, "4", "5", "tz name")
                    ),
                    DurationTuple("6", "7", "8", "9", "10", "11", "12"),
                ),
                {
                    "start": None,
                    "end": TimeTuple(
                        "1", "2", "3", TimezoneTuple(True, False, "4", "5", "tz name")
                    ),
                    "duration": DurationTuple("6", "7", "8", "9", "10", "11", "12"),
                },
            ),
        )

        repeatingintervaltests = (
            (
                RepeatingIntervalTuple(
                    True,
                    None,
                    IntervalTuple(
                        DateTuple("1", "2", "3", "4", "5", "6"),
                        DateTuple("7", "8", "9", "10", "11", "12"),
                        None,
                    ),
                ),
                {
                    "R": True,
                    "Rnn": None,
                    "interval": IntervalTuple(
                        DateTuple("1", "2", "3", "4", "5", "6"),
                        DateTuple("7", "8", "9", "10", "11", "12"),
                        None,
                    ),
                },
            ),
            (
                RepeatingIntervalTuple(
                    False,
                    "1",
                    IntervalTuple(
                        DatetimeTuple(
                            DateTuple("2", "3", "4", "5", "6", "7"),
                            TimeTuple("8", "9", "10", None),
                        ),
                        DatetimeTuple(
                            DateTuple("11", "12", "13", "14", "15", "16"),
                            TimeTuple("17", "18", "19", None),
                        ),
                        None,
                    ),
                ),
                {
                    "R": False,
                    "Rnn": "1",
                    "interval": IntervalTuple(
                        DatetimeTuple(
                            DateTuple("2", "3", "4", "5", "6", "7"),
                            TimeTuple("8", "9", "10", None),
                        ),
                        DatetimeTuple(
                            DateTuple("11", "12", "13", "14", "15", "16"),
                            TimeTuple("17", "18", "19", None),
                        ),
                        None,
                    ),
                },
            ),
        )

        timezonetest = (
            TimezoneTuple(False, False, "1", "2", "+01:02"),
            {"negative": False, "Z": False, "hh": "1", "mm": "2", "name": "+01:02"},
        )

        with mock.patch.object(
            aniso8601.builders.BaseTimeBuilder, "build_date"
        ) as mock_build:
            mock_build.return_value = datetest[0]

            result = BaseTimeBuilder._build_object(datetest[0])

            self.assertEqual(result, datetest[0])
            mock_build.assert_called_once_with(**datetest[1])

        with mock.patch.object(
            aniso8601.builders.BaseTimeBuilder, "build_time"
        ) as mock_build:
            mock_build.return_value = timetest[0]

            result = BaseTimeBuilder._build_object(timetest[0])

            self.assertEqual(result, timetest[0])
            mock_build.assert_called_once_with(**timetest[1])

        with mock.patch.object(
            aniso8601.builders.BaseTimeBuilder, "build_datetime"
        ) as mock_build:
            mock_build.return_value = datetimetest[0]

            result = BaseTimeBuilder._build_object(datetimetest[0])

            self.assertEqual(result, datetimetest[0])
            mock_build.assert_called_once_with(*datetimetest[1])

        with mock.patch.object(
            aniso8601.builders.BaseTimeBuilder, "build_duration"
        ) as mock_build:
            mock_build.return_value = durationtest[0]

            result = BaseTimeBuilder._build_object(durationtest[0])

            self.assertEqual(result, durationtest[0])
            mock_build.assert_called_once_with(**durationtest[1])

        for intervaltest in intervaltests:
            with mock.patch.object(
                aniso8601.builders.BaseTimeBuilder, "build_interval"
            ) as mock_build:
                mock_build.return_value = intervaltest[0]

                result = BaseTimeBuilder._build_object(intervaltest[0])

                self.assertEqual(result, intervaltest[0])
                mock_build.assert_called_once_with(**intervaltest[1])

        for repeatingintervaltest in repeatingintervaltests:
            with mock.patch.object(
                aniso8601.builders.BaseTimeBuilder, "build_repeating_interval"
            ) as mock_build:
                mock_build.return_value = repeatingintervaltest[0]

                result = BaseTimeBuilder._build_object(repeatingintervaltest[0])

                self.assertEqual(result, repeatingintervaltest[0])
                mock_build.assert_called_once_with(**repeatingintervaltest[1])

        with mock.patch.object(
            aniso8601.builders.BaseTimeBuilder, "build_timezone"
        ) as mock_build:
            mock_build.return_value = timezonetest[0]

            result = BaseTimeBuilder._build_object(timezonetest[0])

            self.assertEqual(result, timezonetest[0])
            mock_build.assert_called_once_with(**timezonetest[1])

    def test_is_interval_end_concise(self):
        self.assertTrue(
            BaseTimeBuilder._is_interval_end_concise(TimeTuple("1", "2", "3", None))
        )
        self.assertTrue(
            BaseTimeBuilder._is_interval_end_concise(
                DateTuple(None, "2", "3", "4", "5", "6")
            )
        )
        self.assertTrue(
            BaseTimeBuilder._is_interval_end_concise(
                DatetimeTuple(
                    DateTuple(None, "2", "3", "4", "5", "6"),
                    TimeTuple("7", "8", "9", None),
                )
            )
        )

        self.assertFalse(
            BaseTimeBuilder._is_interval_end_concise(
                DateTuple("1", "2", "3", "4", "5", "6")
            )
        )
        self.assertFalse(
            BaseTimeBuilder._is_interval_end_concise(
                DatetimeTuple(
                    DateTuple("1", "2", "3", "4", "5", "6"),
                    TimeTuple("7", "8", "9", None),
                )
            )
        )

    def test_combine_concise_interval_tuples(self):
        testtuples = (
            (
                DateTuple("2020", "01", "01", None, None, None),
                DateTuple(None, None, "02", None, None, None),
                DateTuple("2020", "01", "02", None, None, None),
            ),
            (
                DateTuple("2008", "02", "15", None, None, None),
                DateTuple(None, "03", "14", None, None, None),
                DateTuple("2008", "03", "14", None, None, None),
            ),
            (
                DatetimeTuple(
                    DateTuple("2007", "12", "14", None, None, None),
                    TimeTuple("13", "30", None, None),
                ),
                TimeTuple("15", "30", None, None),
                DatetimeTuple(
                    DateTuple("2007", "12", "14", None, None, None),
                    TimeTuple("15", "30", None, None),
                ),
            ),
            (
                DatetimeTuple(
                    DateTuple("2007", "11", "13", None, None, None),
                    TimeTuple("09", "00", None, None),
                ),
                DatetimeTuple(
                    DateTuple(None, None, "15", None, None, None),
                    TimeTuple("17", "00", None, None),
                ),
                DatetimeTuple(
                    DateTuple("2007", "11", "15", None, None, None),
                    TimeTuple("17", "00", None, None),
                ),
            ),
            (
                DatetimeTuple(
                    DateTuple("2007", "11", "13", None, None, None),
                    TimeTuple("00", "00", None, None),
                ),
                DatetimeTuple(
                    DateTuple(None, None, "16", None, None, None),
                    TimeTuple("00", "00", None, None),
                ),
                DatetimeTuple(
                    DateTuple("2007", "11", "16", None, None, None),
                    TimeTuple("00", "00", None, None),
                ),
            ),
            (
                DatetimeTuple(
                    DateTuple("2007", "11", "13", None, None, None),
                    TimeTuple(
                        "09", "00", None, TimezoneTuple(False, True, None, None, "Z")
                    ),
                ),
                DatetimeTuple(
                    DateTuple(None, None, "15", None, None, None),
                    TimeTuple("17", "00", None, None),
                ),
                DatetimeTuple(
                    DateTuple("2007", "11", "15", None, None, None),
                    TimeTuple(
                        "17", "00", None, TimezoneTuple(False, True, None, None, "Z")
                    ),
                ),
            ),
        )

        for testtuple in testtuples:
            result = BaseTimeBuilder._combine_concise_interval_tuples(
                testtuple[0], testtuple[1]
            )
            self.assertEqual(result, testtuple[2])


class TestTupleBuilder(unittest.TestCase):
    def test_build_date(self):
        datetuple = TupleBuilder.build_date()

        self.assertEqual(datetuple, DateTuple(None, None, None, None, None, None))

        datetuple = TupleBuilder.build_date(
            YYYY="1", MM="2", DD="3", Www="4", D="5", DDD="6"
        )

        self.assertEqual(datetuple, DateTuple("1", "2", "3", "4", "5", "6"))

    def test_build_time(self):
        testtuples = (
            ({}, TimeTuple(None, None, None, None)),
            (
                {"hh": "1", "mm": "2", "ss": "3", "tz": None},
                TimeTuple("1", "2", "3", None),
            ),
            (
                {
                    "hh": "1",
                    "mm": "2",
                    "ss": "3",
                    "tz": TimezoneTuple(False, False, "4", "5", "tz name"),
                },
                TimeTuple(
                    "1", "2", "3", TimezoneTuple(False, False, "4", "5", "tz name")
                ),
            ),
        )

        for testtuple in testtuples:
            self.assertEqual(TupleBuilder.build_time(**testtuple[0]), testtuple[1])

    def test_build_datetime(self):
        testtuples = (
            (
                {
                    "date": DateTuple("1", "2", "3", "4", "5", "6"),
                    "time": TimeTuple("7", "8", "9", None),
                },
                DatetimeTuple(
                    DateTuple("1", "2", "3", "4", "5", "6"),
                    TimeTuple("7", "8", "9", None),
                ),
            ),
            (
                {
                    "date": DateTuple("1", "2", "3", "4", "5", "6"),
                    "time": TimeTuple(
                        "7", "8", "9", TimezoneTuple(True, False, "10", "11", "tz name")
                    ),
                },
                DatetimeTuple(
                    DateTuple("1", "2", "3", "4", "5", "6"),
                    TimeTuple(
                        "7", "8", "9", TimezoneTuple(True, False, "10", "11", "tz name")
                    ),
                ),
            ),
        )

        for testtuple in testtuples:
            self.assertEqual(TupleBuilder.build_datetime(**testtuple[0]), testtuple[1])

    def test_build_duration(self):
        testtuples = (
            ({}, DurationTuple(None, None, None, None, None, None, None)),
            (
                {
                    "PnY": "1",
                    "PnM": "2",
                    "PnW": "3",
                    "PnD": "4",
                    "TnH": "5",
                    "TnM": "6",
                    "TnS": "7",
                },
                DurationTuple("1", "2", "3", "4", "5", "6", "7"),
            ),
        )

        for testtuple in testtuples:
            self.assertEqual(TupleBuilder.build_duration(**testtuple[0]), testtuple[1])

    def test_build_interval(self):
        testtuples = (
            ({}, IntervalTuple(None, None, None)),
            (
                {
                    "start": DateTuple("1", "2", "3", "4", "5", "6"),
                    "end": DateTuple("7", "8", "9", "10", "11", "12"),
                },
                IntervalTuple(
                    DateTuple("1", "2", "3", "4", "5", "6"),
                    DateTuple("7", "8", "9", "10", "11", "12"),
                    None,
                ),
            ),
            (
                {
                    "start": TimeTuple(
                        "1", "2", "3", TimezoneTuple(True, False, "7", "8", "tz name")
                    ),
                    "end": TimeTuple(
                        "4", "5", "6", TimezoneTuple(False, False, "9", "10", "tz name")
                    ),
                },
                IntervalTuple(
                    TimeTuple(
                        "1", "2", "3", TimezoneTuple(True, False, "7", "8", "tz name")
                    ),
                    TimeTuple(
                        "4", "5", "6", TimezoneTuple(False, False, "9", "10", "tz name")
                    ),
                    None,
                ),
            ),
            (
                {
                    "start": DatetimeTuple(
                        DateTuple("1", "2", "3", "4", "5", "6"),
                        TimeTuple(
                            "7",
                            "8",
                            "9",
                            TimezoneTuple(True, False, "10", "11", "tz name"),
                        ),
                    ),
                    "end": DatetimeTuple(
                        DateTuple("12", "13", "14", "15", "16", "17"),
                        TimeTuple(
                            "18",
                            "19",
                            "20",
                            TimezoneTuple(False, False, "21", "22", "tz name"),
                        ),
                    ),
                },
                IntervalTuple(
                    DatetimeTuple(
                        DateTuple("1", "2", "3", "4", "5", "6"),
                        TimeTuple(
                            "7",
                            "8",
                            "9",
                            TimezoneTuple(True, False, "10", "11", "tz name"),
                        ),
                    ),
                    DatetimeTuple(
                        DateTuple("12", "13", "14", "15", "16", "17"),
                        TimeTuple(
                            "18",
                            "19",
                            "20",
                            TimezoneTuple(False, False, "21", "22", "tz name"),
                        ),
                    ),
                    None,
                ),
            ),
            (
                {
                    "start": DateTuple("1", "2", "3", "4", "5", "6"),
                    "end": None,
                    "duration": DurationTuple("7", "8", "9", "10", "11", "12", "13"),
                },
                IntervalTuple(
                    DateTuple("1", "2", "3", "4", "5", "6"),
                    None,
                    DurationTuple("7", "8", "9", "10", "11", "12", "13"),
                ),
            ),
            (
                {
                    "start": None,
                    "end": TimeTuple(
                        "1", "2", "3", TimezoneTuple(True, False, "4", "5", "tz name")
                    ),
                    "duration": DurationTuple("6", "7", "8", "9", "10", "11", "12"),
                },
                IntervalTuple(
                    None,
                    TimeTuple(
                        "1", "2", "3", TimezoneTuple(True, False, "4", "5", "tz name")
                    ),
                    DurationTuple("6", "7", "8", "9", "10", "11", "12"),
                ),
            ),
        )

        for testtuple in testtuples:
            self.assertEqual(TupleBuilder.build_interval(**testtuple[0]), testtuple[1])

    def test_build_repeating_interval(self):
        testtuples = (
            ({}, RepeatingIntervalTuple(None, None, None)),
            (
                {
                    "R": True,
                    "interval": IntervalTuple(
                        DateTuple("1", "2", "3", "4", "5", "6"),
                        DateTuple("7", "8", "9", "10", "11", "12"),
                        None,
                    ),
                },
                RepeatingIntervalTuple(
                    True,
                    None,
                    IntervalTuple(
                        DateTuple("1", "2", "3", "4", "5", "6"),
                        DateTuple("7", "8", "9", "10", "11", "12"),
                        None,
                    ),
                ),
            ),
            (
                {
                    "R": False,
                    "Rnn": "1",
                    "interval": IntervalTuple(
                        DatetimeTuple(
                            DateTuple("2", "3", "4", "5", "6", "7"),
                            TimeTuple("8", "9", "10", None),
                        ),
                        DatetimeTuple(
                            DateTuple("11", "12", "13", "14", "15", "16"),
                            TimeTuple("17", "18", "19", None),
                        ),
                        None,
                    ),
                },
                RepeatingIntervalTuple(
                    False,
                    "1",
                    IntervalTuple(
                        DatetimeTuple(
                            DateTuple("2", "3", "4", "5", "6", "7"),
                            TimeTuple("8", "9", "10", None),
                        ),
                        DatetimeTuple(
                            DateTuple("11", "12", "13", "14", "15", "16"),
                            TimeTuple("17", "18", "19", None),
                        ),
                        None,
                    ),
                ),
            ),
        )

        for testtuple in testtuples:
            result = TupleBuilder.build_repeating_interval(**testtuple[0])
            self.assertEqual(result, testtuple[1])

    def test_build_timezone(self):
        testtuples = (
            ({}, TimezoneTuple(None, None, None, None, "")),
            (
                {"negative": False, "Z": True, "name": "UTC"},
                TimezoneTuple(False, True, None, None, "UTC"),
            ),
            (
                {"negative": False, "Z": False, "hh": "1", "mm": "2", "name": "+01:02"},
                TimezoneTuple(False, False, "1", "2", "+01:02"),
            ),
            (
                {"negative": True, "Z": False, "hh": "1", "mm": "2", "name": "-01:02"},
                TimezoneTuple(True, False, "1", "2", "-01:02"),
            ),
        )

        for testtuple in testtuples:
            result = TupleBuilder.build_timezone(**testtuple[0])
            self.assertEqual(result, testtuple[1])



# -*- coding: utf-8 -*-

# Copyright (c) 2025, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import unittest

import aniso8601
from aniso8601.builders import (
    DatetimeTuple,
    DateTuple,
    DurationTuple,
    IntervalTuple,
    TimeTuple,
    TimezoneTuple,
)
from aniso8601.exceptions import ISOFormatError
from aniso8601.interval import (
    _get_interval_component_resolution,
    _get_interval_resolution,
    _parse_interval,
    _parse_interval_end,
    get_interval_resolution,
    get_repeating_interval_resolution,
    parse_interval,
    parse_repeating_interval,
)
from aniso8601.resolution import IntervalResolution
from aniso8601.tests.compat import mock


class TestIntervalParser_UtilityFunctions(unittest.TestCase):
    def test_get_interval_resolution(self):
        self.assertEqual(
            _get_interval_resolution(
                IntervalTuple(
                    start=DateTuple(
                        YYYY="2001", MM="02", DD="03", Www=None, D=None, DDD=None
                    ),
                    end=DatetimeTuple(
                        DateTuple(
                            YYYY="2001", MM="02", DD="03", Www=None, D=None, DDD=None
                        ),
                        TimeTuple(hh="04", mm="05", ss="06", tz=None),
                    ),
                    duration=None,
                )
            ),
            IntervalResolution.Seconds,
        )
        self.assertEqual(
            _get_interval_resolution(
                IntervalTuple(
                    start=DatetimeTuple(
                        DateTuple(
                            YYYY="2001", MM="02", DD="03", Www=None, D=None, DDD=None
                        ),
                        TimeTuple(hh="04", mm="05", ss="06", tz=None),
                    ),
                    end=DateTuple(
                        YYYY="2001", MM="02", DD="03", Www=None, D=None, DDD=None
                    ),
                    duration=None,
                )
            ),
            IntervalResolution.Seconds,
        )

        self.assertEqual(
            _get_interval_resolution(
                IntervalTuple(
                    start=DateTuple(
                        YYYY="2001", MM="02", DD="03", Www=None, D=None, DDD=None
                    ),
                    end=None,
                    duration=DurationTuple(
                        PnY="1", PnM="2", PnW=None, PnD="3", TnH="4", TnM="5", TnS="6"
                    ),
                )
            ),
            IntervalResolution.Seconds,
        )
        self.assertEqual(
            _get_interval_resolution(
                IntervalTuple(
                    start=DatetimeTuple(
                        DateTuple(
                            YYYY="2001", MM="02", DD="03", Www=None, D=None, DDD=None
                        ),
                        TimeTuple(hh="04", mm="05", ss="06", tz=None),
                    ),
                    end=None,
                    duration=DurationTuple(
                        PnY="1",
                        PnM="2",
                        PnW=None,
                        PnD="3",
                        TnH=None,
                        TnM=None,
                        TnS=None,
                    ),
                )
            ),
            IntervalResolution.Seconds,
        )

        self.assertEqual(
            _get_interval_resolution(
                IntervalTuple(
                    start=None,
                    end=DateTuple(
                        YYYY="2001", MM="02", DD="03", Www=None, D=None, DDD=None
                    ),
                    duration=DurationTuple(
                        PnY="1", PnM="2", PnW=None, PnD="3", TnH="4", TnM="5", TnS="6"
                    ),
                )
            ),
            IntervalResolution.Seconds,
        )
        self.assertEqual(
            _get_interval_resolution(
                IntervalTuple(
                    start=None,
                    end=DatetimeTuple(
                        DateTuple(
                            YYYY="2001", MM="02", DD="03", Www=None, D=None, DDD=None
                        ),
                        TimeTuple(hh="04", mm="05", ss="06", tz=None),
                    ),
                    duration=DurationTuple(
                        PnY="1",
                        PnM="2",
                        PnW=None,
                        PnD="3",
                        TnH=None,
                        TnM=None,
                        TnS=None,
                    ),
                )
            ),
            IntervalResolution.Seconds,
        )

    def test_get_interval_component_resolution(self):
        self.assertEqual(
            _get_interval_component_resolution(
                DateTuple(YYYY="2001", MM=None, DD=None, Www=None, D=None, DDD="123")
            ),
            IntervalResolution.Ordinal,
        )
        self.assertEqual(
            _get_interval_component_resolution(
                DateTuple(YYYY="2001", MM=None, DD=None, Www="12", D="3", DDD=None)
            ),
            IntervalResolution.Weekday,
        )
        self.assertEqual(
            _get_interval_component_resolution(
                DateTuple(YYYY="2001", MM=None, DD=None, Www="12", D=None, DDD=None)
            ),
            IntervalResolution.Week,
        )
        self.assertEqual(
            _get_interval_component_resolution(
                DateTuple(YYYY="2001", MM="02", DD="03", Www=None, D=None, DDD=None)
            ),
            IntervalResolution.Day,
        )
        self.assertEqual(
            _get_interval_component_resolution(
                DateTuple(YYYY="2001", MM="02", DD=None, Www=None, D=None, DDD=None)
            ),
            IntervalResolution.Month,
        )
        self.assertEqual(
            _get_interval_component_resolution(
                DateTuple(YYYY="2001", MM=None, DD=None, Www=None, D=None, DDD=None)
            ),
            IntervalResolution.Year,
        )

        self.assertEqual(
            _get_interval_component_resolution(
                DatetimeTuple(
                    DateTuple(
                        YYYY="2001", MM="02", DD="03", Www=None, D=None, DDD=None
                    ),
                    TimeTuple(hh="04", mm="05", ss="06", tz=None),
                )
            ),
            IntervalResolution.Seconds,
        )
        self.assertEqual(
            _get_interval_component_resolution(
                DatetimeTuple(
                    DateTuple(
                        YYYY="2001", MM="02", DD="03", Www=None, D=None, DDD=None
                    ),
                    TimeTuple(hh="04", mm="05", ss=None, tz=None),
                )
            ),
            IntervalResolution.Minutes,
        )
        self.assertEqual(
            _get_interval_component_resolution(
                DatetimeTuple(
                    DateTuple(
                        YYYY="2001", MM="02", DD="03", Www=None, D=None, DDD=None
                    ),
                    TimeTuple(hh="04", mm=None, ss=None, tz=None),
                )
            ),
            IntervalResolution.Hours,
        )

        self.assertEqual(
            _get_interval_component_resolution(
                DurationTuple(
                    PnY="1", PnM="2", PnW=None, PnD="3", TnH="4", TnM="5", TnS="6"
                )
            ),
            IntervalResolution.Seconds,
        )
        self.assertEqual(
            _get_interval_component_resolution(
                DurationTuple(
                    PnY="1", PnM="2", PnW=None, PnD="3", TnH="4", TnM="5", TnS=None
                )
            ),
            IntervalResolution.Minutes,
        )
        self.assertEqual(
            _get_interval_component_resolution(
                DurationTuple(
                    PnY="1", PnM="2", PnW=None, PnD="3", TnH="4", TnM=None, TnS=None
                )
            ),
            IntervalResolution.Hours,
        )
        self.assertEqual(
            _get_interval_component_resolution(
                DurationTuple(
                    PnY="1", PnM="2", PnW=None, PnD="3", TnH=None, TnM=None, TnS=None
                )
            ),
            IntervalResolution.Day,
        )
        self.assertEqual(
            _get_interval_component_resolution(
                DurationTuple(
                    PnY="1", PnM="2", PnW=None, PnD=None, TnH=None, TnM=None, TnS=None
                )
            ),
            IntervalResolution.Month,
        )
        self.assertEqual(
            _get_interval_component_resolution(
                DurationTuple(
                    PnY="1", PnM=None, PnW=None, PnD=None, TnH=None, TnM=None, TnS=None
                )
            ),
            IntervalResolution.Year,
        )
        self.assertEqual(
            _get_interval_component_resolution(
                DurationTuple(
                    PnY=None, PnM=None, PnW="3", PnD=None, TnH=None, TnM=None, TnS=None
                )
            ),
            IntervalResolution.Week,
        )


class TestIntervalParserFunctions(unittest.TestCase):
    def test_get_interval_resolution_date(self):
        self.assertEqual(get_interval_resolution("P1.5Y/2018"), IntervalResolution.Year)
        self.assertEqual(
            get_interval_resolution("P1.5Y/2018-03"), IntervalResolution.Month
        )
        self.assertEqual(
            get_interval_resolution("P1.5Y/2018-03-06"), IntervalResolution.Day
        )
        self.assertEqual(
            get_interval_resolution("P1.5Y/2018W01"), IntervalResolution.Week
        )
        self.assertEqual(
            get_interval_resolution("P1.5Y/2018-306"), IntervalResolution.Ordinal
        )
        self.assertEqual(
            get_interval_resolution("P1.5Y/2018W012"), IntervalResolution.Weekday
        )

        self.assertEqual(get_interval_resolution("2018/P1.5Y"), IntervalResolution.Year)
        self.assertEqual(
            get_interval_resolution("2018-03/P1.5Y"), IntervalResolution.Month
        )
        self.assertEqual(
            get_interval_resolution("2018-03-06/P1.5Y"), IntervalResolution.Day
        )
        self.assertEqual(
            get_interval_resolution("2018W01/P1.5Y"), IntervalResolution.Week
        )
        self.assertEqual(
            get_interval_resolution("2018-306/P1.5Y"), IntervalResolution.Ordinal
        )
        self.assertEqual(
            get_interval_resolution("2018W012/P1.5Y"), IntervalResolution.Weekday
        )

    def test_get_interval_resolution_time(self):
        self.assertEqual(
            get_interval_resolution("P1M/1981-04-05T01"), IntervalResolution.Hours
        )
        self.assertEqual(
            get_interval_resolution("P1M/1981-04-05T01:01"), IntervalResolution.Minutes
        )
        self.assertEqual(
            get_interval_resolution("P1M/1981-04-05T01:01:00"),
            IntervalResolution.Seconds,
        )

        self.assertEqual(
            get_interval_resolution("1981-04-05T01/P1M"), IntervalResolution.Hours
        )
        self.assertEqual(
            get_interval_resolution("1981-04-05T01:01/P1M"), IntervalResolution.Minutes
        )
        self.assertEqual(
            get_interval_resolution("1981-04-05T01:01:00/P1M"),
            IntervalResolution.Seconds,
        )

    def test_get_interval_resolution_duration(self):
        self.assertEqual(
            get_interval_resolution("2014-11-12/P1Y2M3D"), IntervalResolution.Day
        )
        self.assertEqual(
            get_interval_resolution("2014-11-12/P1Y2M"), IntervalResolution.Day
        )
        self.assertEqual(
            get_interval_resolution("2014-11-12/P1Y"), IntervalResolution.Day
        )
        self.assertEqual(
            get_interval_resolution("2014-11-12/P1W"), IntervalResolution.Day
        )
        self.assertEqual(
            get_interval_resolution("2014-11-12/P1Y2M3DT4H"), IntervalResolution.Hours
        )
        self.assertEqual(
            get_interval_resolution("2014-11-12/P1Y2M3DT4H54M"),
            IntervalResolution.Minutes,
        )
        self.assertEqual(
            get_interval_resolution("2014-11-12/P1Y2M3DT4H54M6S"),
            IntervalResolution.Seconds,
        )

        self.assertEqual(
            get_interval_resolution("P1Y2M3D/2014-11-12"), IntervalResolution.Day
        )
        self.assertEqual(
            get_interval_resolution("P1Y2M/2014-11-12"), IntervalResolution.Day
        )
        self.assertEqual(
            get_interval_resolution("P1Y/2014-11-12"), IntervalResolution.Day
        )
        self.assertEqual(
            get_interval_resolution("P1W/2014-11-12"), IntervalResolution.Day
        )
        self.assertEqual(
            get_interval_resolution("P1Y2M3DT4H/2014-11-12"), IntervalResolution.Hours
        )
        self.assertEqual(
            get_interval_resolution("P1Y2M3DT4H54M/2014-11-12"),
            IntervalResolution.Minutes,
        )
        self.assertEqual(
            get_interval_resolution("P1Y2M3DT4H54M6S/2014-11-12"),
            IntervalResolution.Seconds,
        )

    def test_parse_interval(self):
        testtuples = (
            (
                "P1M/1981-04-05T01:01:00",
                {
                    "end": DatetimeTuple(
                        DateTuple("1981", "04", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                    "duration": DurationTuple(None, "1", None, None, None, None, None),
                },
            ),
            (
                "P1M/1981-04-05",
                {
                    "end": DateTuple("1981", "04", "05", None, None, None),
                    "duration": DurationTuple(None, "1", None, None, None, None, None),
                },
            ),
            (
                "P1,5Y/2018-03-06",
                {
                    "end": DateTuple("2018", "03", "06", None, None, None),
                    "duration": DurationTuple(
                        "1.5", None, None, None, None, None, None
                    ),
                },
            ),
            (
                "P1.5Y/2018-03-06",
                {
                    "end": DateTuple("2018", "03", "06", None, None, None),
                    "duration": DurationTuple(
                        "1.5", None, None, None, None, None, None
                    ),
                },
            ),
            (
                "PT1H/2014-11-12",
                {
                    "end": DateTuple("2014", "11", "12", None, None, None),
                    "duration": DurationTuple(None, None, None, None, "1", None, None),
                },
            ),
            (
                "PT4H54M6.5S/2014-11-12",
                {
                    "end": DateTuple("2014", "11", "12", None, None, None),
                    "duration": DurationTuple(None, None, None, None, "4", "54", "6.5"),
                },
            ),
            (
                "PT10H/2050-03-01T13:00:00Z",
                {
                    "end": DatetimeTuple(
                        DateTuple("2050", "03", "01", None, None, None),
                        TimeTuple(
                            "13",
                            "00",
                            "00",
                            TimezoneTuple(False, True, None, None, "Z"),
                        ),
                    ),
                    "duration": DurationTuple(None, None, None, None, "10", None, None),
                },
            ),
            # Make sure we truncate, not round
            # https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
            (
                "PT0.0000001S/2018-03-06",
                {
                    "end": DateTuple("2018", "03", "06", None, None, None),
                    "duration": DurationTuple(
                        None, None, None, None, None, None, "0.0000001"
                    ),
                },
            ),
            (
                "PT2.0000048S/2018-03-06",
                {
                    "end": DateTuple("2018", "03", "06", None, None, None),
                    "duration": DurationTuple(
                        None, None, None, None, None, None, "2.0000048"
                    ),
                },
            ),
            (
                "1981-04-05T01:01:00/P1M1DT1M",
                {
                    "start": DatetimeTuple(
                        DateTuple("1981", "04", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                    "duration": DurationTuple(None, "1", None, "1", None, "1", None),
                },
            ),
            (
                "1981-04-05/P1M1D",
                {
                    "start": DateTuple("1981", "04", "05", None, None, None),
                    "duration": DurationTuple(None, "1", None, "1", None, None, None),
                },
            ),
            (
                "2018-03-06/P2,5M",
                {
                    "start": DateTuple("2018", "03", "06", None, None, None),
                    "duration": DurationTuple(
                        None, "2.5", None, None, None, None, None
                    ),
                },
            ),
            (
                "2018-03-06/P2.5M",
                {
                    "start": DateTuple("2018", "03", "06", None, None, None),
                    "duration": DurationTuple(
                        None, "2.5", None, None, None, None, None
                    ),
                },
            ),
            (
                "2014-11-12/PT1H",
                {
                    "start": DateTuple("2014", "11", "12", None, None, None),
                    "duration": DurationTuple(None, None, None, None, "1", None, None),
                },
            ),
            (
                "2014-11-12/PT4H54M6.5S",
                {
                    "start": DateTuple("2014", "11", "12", None, None, None),
                    "duration": DurationTuple(None, None, None, None, "4", "54", "6.5"),
                },
            ),
            (
                "2050-03-01T13:00:00Z/PT10H",
                {
                    "start": DatetimeTuple(
                        DateTuple("2050", "03", "01", None, None, None),
                        TimeTuple(
                            "13",
                            "00",
                            "00",
                            TimezoneTuple(False, True, None, None, "Z"),
                        ),
                    ),
                    "duration": DurationTuple(None, None, None, None, "10", None, None),
                },
            ),
            # Make sure we truncate, not round
            # https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
            (
                "2018-03-06/PT0.0000001S",
                {
                    "start": DateTuple("2018", "03", "06", None, None, None),
                    "duration": DurationTuple(
                        None, None, None, None, None, None, "0.0000001"
                    ),
                },
            ),
            (
                "2018-03-06/PT2.0000048S",
                {
                    "start": DateTuple("2018", "03", "06", None, None, None),
                    "duration": DurationTuple(
                        None, None, None, None, None, None, "2.0000048"
                    ),
                },
            ),
            (
                "1980-03-05T01:01:00/1981-04-05T01:01:00",
                {
                    "start": DatetimeTuple(
                        DateTuple("1980", "03", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                    "end": DatetimeTuple(
                        DateTuple("1981", "04", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                },
            ),
            (
                "1980-03-05T01:01:00/1981-04-05",
                {
                    "start": DatetimeTuple(
                        DateTuple("1980", "03", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                    "end": DateTuple("1981", "04", "05", None, None, None),
                },
            ),
            (
                "1980-03-05/1981-04-05T01:01:00",
                {
                    "start": DateTuple("1980", "03", "05", None, None, None),
                    "end": DatetimeTuple(
                        DateTuple("1981", "04", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                },
            ),
            (
                "1980-03-05/1981-04-05",
                {
                    "start": DateTuple("1980", "03", "05", None, None, None),
                    "end": DateTuple("1981", "04", "05", None, None, None),
                },
            ),
            (
                "1981-04-05/1980-03-05",
                {
                    "start": DateTuple("1981", "04", "05", None, None, None),
                    "end": DateTuple("1980", "03", "05", None, None, None),
                },
            ),
            (
                "2050-03-01T13:00:00Z/2050-05-11T15:30:00Z",
                {
                    "start": DatetimeTuple(
                        DateTuple("2050", "03", "01", None, None, None),
                        TimeTuple(
                            "13",
                            "00",
                            "00",
                            TimezoneTuple(False, True, None, None, "Z"),
                        ),
                    ),
                    "end": DatetimeTuple(
                        DateTuple("2050", "05", "11", None, None, None),
                        TimeTuple(
                            "15",
                            "30",
                            "00",
                            TimezoneTuple(False, True, None, None, "Z"),
                        ),
                    ),
                },
            ),
            # Test concise interval
            (
                "2020-01-01/02",
                {
                    "start": DateTuple("2020", "01", "01", None, None, None),
                    "end": DateTuple(None, None, "02", None, None, None),
                },
            ),
            (
                "2008-02-15/03-14",
                {
                    "start": DateTuple("2008", "02", "15", None, None, None),
                    "end": DateTuple(None, "03", "14", None, None, None),
                },
            ),
            (
                "2007-12-14T13:30/15:30",
                {
                    "start": DatetimeTuple(
                        DateTuple("2007", "12", "14", None, None, None),
                        TimeTuple("13", "30", None, None),
                    ),
                    "end": TimeTuple("15", "30", None, None),
                },
            ),
            (
                "2007-11-13T09:00/15T17:00",
                {
                    "start": DatetimeTuple(
                        DateTuple("2007", "11", "13", None, None, None),
                        TimeTuple("09", "00", None, None),
                    ),
                    "end": DatetimeTuple(
                        DateTuple(None, None, "15", None, None, None),
                        TimeTuple("17", "00", None, None),
                    ),
                },
            ),
            (
                "2007-11-13T00:00/16T00:00",
                {
                    "start": DatetimeTuple(
                        DateTuple("2007", "11", "13", None, None, None),
                        TimeTuple("00", "00", None, None),
                    ),
                    "end": DatetimeTuple(
                        DateTuple(None, None, "16", None, None, None),
                        TimeTuple("00", "00", None, None),
                    ),
                },
            ),
            (
                "2007-11-13T09:00Z/15T17:00",
                {
                    "start": DatetimeTuple(
                        DateTuple("2007", "11", "13", None, None, None),
                        TimeTuple(
                            "09",
                            "00",
                            None,
                            TimezoneTuple(False, True, None, None, "Z"),
                        ),
                    ),
                    "end": DatetimeTuple(
                        DateTuple(None, None, "15", None, None, None),
                        TimeTuple("17", "00", None, None),
                    ),
                },
            ),
            (
                "2007-11-13T00:00/12:34.567",
                {
                    "start": DatetimeTuple(
                        DateTuple("2007", "11", "13", None, None, None),
                        TimeTuple("00", "00", None, None),
                    ),
                    "end": TimeTuple("12", "34.567", None, None),
                },
            ),
            # Make sure we truncate, not round
            # https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
            (
                "1980-03-05T01:01:00.0000001/" "1981-04-05T14:43:59.9999997",
                {
                    "start": DatetimeTuple(
                        DateTuple("1980", "03", "05", None, None, None),
                        TimeTuple("01", "01", "00.0000001", None),
                    ),
                    "end": DatetimeTuple(
                        DateTuple("1981", "04", "05", None, None, None),
                        TimeTuple("14", "43", "59.9999997", None),
                    ),
                },
            ),
        )

        for testtuple in testtuples:
            with mock.patch.object(
                aniso8601.interval.PythonTimeBuilder, "build_interval"
            ) as mockBuildInterval:
                mockBuildInterval.return_value = testtuple[1]

                result = parse_interval(testtuple[0])

                self.assertEqual(result, testtuple[1])
                mockBuildInterval.assert_called_once_with(**testtuple[1])

        # Test different separators
        with mock.patch.object(
            aniso8601.interval.PythonTimeBuilder, "build_interval"
        ) as mockBuildInterval:
            expectedargs = {
                "start": DatetimeTuple(
                    DateTuple("1980", "03", "05", None, None, None),
                    TimeTuple("01", "01", "00", None),
                ),
                "end": DatetimeTuple(
                    DateTuple("1981", "04", "05", None, None, None),
                    TimeTuple("01", "01", "00", None),
                ),
            }

            mockBuildInterval.return_value = expectedargs

            result = parse_interval(
                "1980-03-05T01:01:00--1981-04-05T01:01:00", intervaldelimiter="--"
            )

            self.assertEqual(result, expectedargs)
            mockBuildInterval.assert_called_once_with(**expectedargs)

        with mock.patch.object(
            aniso8601.interval.PythonTimeBuilder, "build_interval"
        ) as mockBuildInterval:
            expectedargs = {
                "start": DatetimeTuple(
                    DateTuple("1980", "03", "05", None, None, None),
                    TimeTuple("01", "01", "00", None),
                ),
                "end": DatetimeTuple(
                    DateTuple("1981", "04", "05", None, None, None),
                    TimeTuple("01", "01", "00", None),
                ),
            }

            mockBuildInterval.return_value = expectedargs

            result = parse_interval(
                "1980-03-05 01:01:00/1981-04-05 01:01:00", datetimedelimiter=" "
            )

            self.assertEqual(result, expectedargs)
            mockBuildInterval.assert_called_once_with(**expectedargs)

    def test_parse_interval_mockbuilder(self):
        mockBuilder = mock.Mock()

        expectedargs = {
            "end": DatetimeTuple(
                DateTuple("1981", "04", "05", None, None, None),
                TimeTuple("01", "01", "00", None),
            ),
            "duration": DurationTuple(None, "1", None, None, None, None, None),
        }

        mockBuilder.build_interval.return_value = expectedargs

        result = parse_interval("P1M/1981-04-05T01:01:00", builder=mockBuilder)

        self.assertEqual(result, expectedargs)
        mockBuilder.build_interval.assert_called_once_with(**expectedargs)

        mockBuilder = mock.Mock()

        expectedargs = {
            "start": DateTuple("2014", "11", "12", None, None, None),
            "duration": DurationTuple(None, None, None, None, "1", None, None),
        }

        mockBuilder.build_interval.return_value = expectedargs

        result = parse_interval("2014-11-12/PT1H", builder=mockBuilder)

        self.assertEqual(result, expectedargs)
        mockBuilder.build_interval.assert_called_once_with(**expectedargs)

        mockBuilder = mock.Mock()

        expectedargs = {
            "start": DatetimeTuple(
                DateTuple("1980", "03", "05", None, None, None),
                TimeTuple("01", "01", "00", None),
            ),
            "end": DatetimeTuple(
                DateTuple("1981", "04", "05", None, None, None),
                TimeTuple("01", "01", "00", None),
            ),
        }

        mockBuilder.build_interval.return_value = expectedargs

        result = parse_interval(
            "1980-03-05T01:01:00/1981-04-05T01:01:00", builder=mockBuilder
        )

        self.assertEqual(result, expectedargs)
        mockBuilder.build_interval.assert_called_once_with(**expectedargs)

    def test_parse_interval_badtype(self):
        testtuples = (None, 1, False, 1.234)

        for testtuple in testtuples:
            with self.assertRaises(ValueError):
                parse_interval(testtuple, builder=None)

    def test_parse_interval_baddelimiter(self):
        testtuples = (
            "1980-03-05T01:01:00,1981-04-05T01:01:00",
            "P1M 1981-04-05T01:01:00",
        )

        for testtuple in testtuples:
            with self.assertRaises(ISOFormatError):
                parse_interval(testtuple, builder=None)

    def test_parse_interval_badstr(self):
        testtuples = ("/", "0/0/0", "20.50230/0", "5/%", "1/21", "bad", "")

        for testtuple in testtuples:
            with self.assertRaises(ISOFormatError):
                parse_interval(testtuple, builder=None)

    def test_parse_interval_repeating(self):
        # Parse interval can't parse repeating intervals
        with self.assertRaises(ISOFormatError):
            parse_interval("R3/1981-04-05/P1D")

        with self.assertRaises(ISOFormatError):
            parse_interval("R3/1981-04-05/P0003-06-04T12:30:05.5")

        with self.assertRaises(ISOFormatError):
            parse_interval("R/PT1H2M/1980-03-05T01:01:00")

    def test_parse_interval_suffixgarbage(self):
        # Don't allow garbage after the duration
        # https://bitbucket.org/nielsenb/aniso8601/issues/9/durations-with-trailing-garbage-are-parsed
        with self.assertRaises(ISOFormatError):
            parse_interval("2001/P1Dasdf", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_interval("P1Dasdf/2001", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_interval("2001/P0003-06-04T12:30:05.5asdfasdf", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_interval("P0003-06-04T12:30:05.5asdfasdf/2001", builder=None)

    def test_parse_interval_internal(self):
        # Test the internal _parse_interval function
        testtuples = (
            (
                "P1M/1981-04-05T01:01:00",
                {
                    "end": DatetimeTuple(
                        DateTuple("1981", "04", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                    "duration": DurationTuple(None, "1", None, None, None, None, None),
                },
            ),
            (
                "P1M/1981-04-05",
                {
                    "end": DateTuple("1981", "04", "05", None, None, None),
                    "duration": DurationTuple(None, "1", None, None, None, None, None),
                },
            ),
            (
                "P1,5Y/2018-03-06",
                {
                    "end": DateTuple("2018", "03", "06", None, None, None),
                    "duration": DurationTuple(
                        "1.5", None, None, None, None, None, None
                    ),
                },
            ),
            (
                "P1.5Y/2018-03-06",
                {
                    "end": DateTuple("2018", "03", "06", None, None, None),
                    "duration": DurationTuple(
                        "1.5", None, None, None, None, None, None
                    ),
                },
            ),
            (
                "PT1H/2014-11-12",
                {
                    "end": DateTuple("2014", "11", "12", None, None, None),
                    "duration": DurationTuple(None, None, None, None, "1", None, None),
                },
            ),
            (
                "PT4H54M6.5S/2014-11-12",
                {
                    "end": DateTuple("2014", "11", "12", None, None, None),
                    "duration": DurationTuple(None, None, None, None, "4", "54", "6.5"),
                },
            ),
            # Make sure we truncate, not round
            # https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
            (
                "PT0.0000001S/2018-03-06",
                {
                    "end": DateTuple("2018", "03", "06", None, None, None),
                    "duration": DurationTuple(
                        None, None, None, None, None, None, "0.0000001"
                    ),
                },
            ),
            (
                "PT2.0000048S/2018-03-06",
                {
                    "end": DateTuple("2018", "03", "06", None, None, None),
                    "duration": DurationTuple(
                        None, None, None, None, None, None, "2.0000048"
                    ),
                },
            ),
            (
                "1981-04-05T01:01:00/P1M1DT1M",
                {
                    "start": DatetimeTuple(
                        DateTuple("1981", "04", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                    "duration": DurationTuple(None, "1", None, "1", None, "1", None),
                },
            ),
            (
                "1981-04-05/P1M1D",
                {
                    "start": DateTuple("1981", "04", "05", None, None, None),
                    "duration": DurationTuple(None, "1", None, "1", None, None, None),
                },
            ),
            (
                "2018-03-06/P2,5M",
                {
                    "start": DateTuple("2018", "03", "06", None, None, None),
                    "duration": DurationTuple(
                        None, "2.5", None, None, None, None, None
                    ),
                },
            ),
            (
                "2018-03-06/P2.5M",
                {
                    "start": DateTuple("2018", "03", "06", None, None, None),
                    "duration": DurationTuple(
                        None, "2.5", None, None, None, None, None
                    ),
                },
            ),
            (
                "2014-11-12/PT1H",
                {
                    "start": DateTuple("2014", "11", "12", None, None, None),
                    "duration": DurationTuple(None, None, None, None, "1", None, None),
                },
            ),
            (
                "2014-11-12/PT4H54M6.5S",
                {
                    "start": DateTuple("2014", "11", "12", None, None, None),
                    "duration": DurationTuple(None, None, None, None, "4", "54", "6.5"),
                },
            ),
            # Make sure we truncate, not round
            # https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
            (
                "2018-03-06/PT0.0000001S",
                {
                    "start": DateTuple("2018", "03", "06", None, None, None),
                    "duration": DurationTuple(
                        None, None, None, None, None, None, "0.0000001"
                    ),
                },
            ),
            (
                "2018-03-06/PT2.0000048S",
                {
                    "start": DateTuple("2018", "03", "06", None, None, None),
                    "duration": DurationTuple(
                        None, None, None, None, None, None, "2.0000048"
                    ),
                },
            ),
            (
                "1980-03-05T01:01:00/1981-04-05T01:01:00",
                {
                    "start": DatetimeTuple(
                        DateTuple("1980", "03", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                    "end": DatetimeTuple(
                        DateTuple("1981", "04", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                },
            ),
            (
                "1980-03-05T01:01:00/1981-04-05",
                {
                    "start": DatetimeTuple(
                        DateTuple("1980", "03", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                    "end": DateTuple("1981", "04", "05", None, None, None),
                },
            ),
            (
                "1980-03-05/1981-04-05T01:01:00",
                {
                    "start": DateTuple("1980", "03", "05", None, None, None),
                    "end": DatetimeTuple(
                        DateTuple("1981", "04", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                },
            ),
            (
                "1980-03-05/1981-04-05",
                {
                    "start": DateTuple("1980", "03", "05", None, None, None),
                    "end": DateTuple("1981", "04", "05", None, None, None),
                },
            ),
            (
                "1981-04-05/1980-03-05",
                {
                    "start": DateTuple("1981", "04", "05", None, None, None),
                    "end": DateTuple("1980", "03", "05", None, None, None),
                },
            ),
            # Test concise interval
            (
                "2020-01-01/02",
                {
                    "start": DateTuple("2020", "01", "01", None, None, None),
                    "end": DateTuple(None, None, "02", None, None, None),
                },
            ),
            (
                "2008-02-15/03-14",
                {
                    "start": DateTuple("2008", "02", "15", None, None, None),
                    "end": DateTuple(None, "03", "14", None, None, None),
                },
            ),
            (
                "2007-12-14T13:30/15:30",
                {
                    "start": DatetimeTuple(
                        DateTuple("2007", "12", "14", None, None, None),
                        TimeTuple("13", "30", None, None),
                    ),
                    "end": TimeTuple("15", "30", None, None),
                },
            ),
            (
                "2007-11-13T09:00/15T17:00",
                {
                    "start": DatetimeTuple(
                        DateTuple("2007", "11", "13", None, None, None),
                        TimeTuple("09", "00", None, None),
                    ),
                    "end": DatetimeTuple(
                        DateTuple(None, None, "15", None, None, None),
                        TimeTuple("17", "00", None, None),
                    ),
                },
            ),
            (
                "2007-11-13T00:00/16T00:00",
                {
                    "start": DatetimeTuple(
                        DateTuple("2007", "11", "13", None, None, None),
                        TimeTuple("00", "00", None, None),
                    ),
                    "end": DatetimeTuple(
                        DateTuple(None, None, "16", None, None, None),
                        TimeTuple("00", "00", None, None),
                    ),
                },
            ),
            (
                "2007-11-13T09:00Z/15T17:00",
                {
                    "start": DatetimeTuple(
                        DateTuple("2007", "11", "13", None, None, None),
                        TimeTuple(
                            "09",
                            "00",
                            None,
                            TimezoneTuple(False, True, None, None, "Z"),
                        ),
                    ),
                    "end": DatetimeTuple(
                        DateTuple(None, None, "15", None, None, None),
                        TimeTuple("17", "00", None, None),
                    ),
                },
            ),
            (
                "2007-11-13T00:00/12:34.567",
                {
                    "start": DatetimeTuple(
                        DateTuple("2007", "11", "13", None, None, None),
                        TimeTuple("00", "00", None, None),
                    ),
                    "end": TimeTuple("12", "34.567", None, None),
                },
            ),
            # Make sure we truncate, not round
            # https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
            (
                "1980-03-05T01:01:00.0000001/" "1981-04-05T14:43:59.9999997",
                {
                    "start": DatetimeTuple(
                        DateTuple("1980", "03", "05", None, None, None),
                        TimeTuple("01", "01", "00.0000001", None),
                    ),
                    "end": DatetimeTuple(
                        DateTuple("1981", "04", "05", None, None, None),
                        TimeTuple("14", "43", "59.9999997", None),
                    ),
                },
            ),
        )

        for testtuple in testtuples:
            mockBuilder = mock.Mock()
            mockBuilder.build_interval.return_value = testtuple[1]

            result = _parse_interval(testtuple[0], mockBuilder)

            self.assertEqual(result, testtuple[1])
            mockBuilder.build_interval.assert_called_once_with(**testtuple[1])

        # Test different separators
        expectedargs = {
            "start": DatetimeTuple(
                DateTuple("1980", "03", "05", None, None, None),
                TimeTuple("01", "01", "00", None),
            ),
            "end": DatetimeTuple(
                DateTuple("1981", "04", "05", None, None, None),
                TimeTuple("01", "01", "00", None),
            ),
        }

        mockBuilder = mock.Mock()
        mockBuilder.build_interval.return_value = expectedargs

        result = _parse_interval(
            "1980-03-05T01:01:00--1981-04-05T01:01:00",
            mockBuilder,
            intervaldelimiter="--",
        )

        self.assertEqual(result, expectedargs)
        mockBuilder.build_interval.assert_called_once_with(**expectedargs)

        expectedargs = {
            "start": DatetimeTuple(
                DateTuple("1980", "03", "05", None, None, None),
                TimeTuple("01", "01", "00", None),
            ),
            "end": DatetimeTuple(
                DateTuple("1981", "04", "05", None, None, None),
                TimeTuple("01", "01", "00", None),
            ),
        }

        mockBuilder = mock.Mock()
        mockBuilder.build_interval.return_value = expectedargs

        _parse_interval(
            "1980-03-05 01:01:00/1981-04-05 01:01:00",
            mockBuilder,
            datetimedelimiter=" ",
        )

        self.assertEqual(result, expectedargs)
        mockBuilder.build_interval.assert_called_once_with(**expectedargs)

    def test_parse_interval_end(self):
        self.assertEqual(
            _parse_interval_end(
                "02", DateTuple("2020", "01", "01", None, None, None), "T"
            ),
            DateTuple(None, None, "02", None, None, None),
        )

        self.assertEqual(
            _parse_interval_end(
                "03-14", DateTuple("2008", "02", "15", None, None, None), "T"
            ),
            DateTuple(None, "03", "14", None, None, None),
        )

        self.assertEqual(
            _parse_interval_end(
                "0314", DateTuple("2008", "02", "15", None, None, None), "T"
            ),
            DateTuple(None, "03", "14", None, None, None),
        )

        self.assertEqual(
            _parse_interval_end(
                "15:30",
                DatetimeTuple(
                    DateTuple("2007", "12", "14", None, None, None),
                    TimeTuple("13", "30", None, None),
                ),
                "T",
            ),
            TimeTuple("15", "30", None, None),
        )

        self.assertEqual(
            _parse_interval_end(
                "15T17:00",
                DatetimeTuple(
                    DateTuple("2007", "11", "13", None, None, None),
                    TimeTuple("09", "00", None, None),
                ),
                "T",
            ),
            DatetimeTuple(
                DateTuple(None, None, "15", None, None, None),
                TimeTuple("17", "00", None, None),
            ),
        )

        self.assertEqual(
            _parse_interval_end(
                "16T00:00",
                DatetimeTuple(
                    DateTuple("2007", "11", "13", None, None, None),
                    TimeTuple("00", "00", None, None),
                ),
                "T",
            ),
            DatetimeTuple(
                DateTuple(None, None, "16", None, None, None),
                TimeTuple("00", "00", None, None),
            ),
        )

        self.assertEqual(
            _parse_interval_end(
                "15 17:00",
                DatetimeTuple(
                    DateTuple("2007", "11", "13", None, None, None),
                    TimeTuple("09", "00", None, None),
                ),
                " ",
            ),
            DatetimeTuple(
                DateTuple(None, None, "15", None, None, None),
                TimeTuple("17", "00", None, None),
            ),
        )

        self.assertEqual(
            _parse_interval_end(
                "12:34.567",
                DatetimeTuple(
                    DateTuple("2007", "11", "13", None, None, None),
                    TimeTuple("00", "00", None, None),
                ),
                "T",
            ),
            TimeTuple("12", "34.567", None, None),
        )


class TestRepeatingIntervalParserFunctions(unittest.TestCase):
    def test_get_interval_resolution_date(self):
        self.assertEqual(
            get_repeating_interval_resolution("R/P1.5Y/2018"), IntervalResolution.Year
        )
        self.assertEqual(
            get_repeating_interval_resolution("R1/P1.5Y/2018-03"),
            IntervalResolution.Month,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R2/P1.5Y/2018-03-06"),
            IntervalResolution.Day,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R3/P1.5Y/2018W01"),
            IntervalResolution.Week,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R4/P1.5Y/2018-306"),
            IntervalResolution.Ordinal,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R5/P1.5Y/2018W012"),
            IntervalResolution.Weekday,
        )

        self.assertEqual(
            get_repeating_interval_resolution("R/2018/P1.5Y"), IntervalResolution.Year
        )
        self.assertEqual(
            get_repeating_interval_resolution("R1/2018-03/P1.5Y"),
            IntervalResolution.Month,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R2/2018-03-06/P1.5Y"),
            IntervalResolution.Day,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R3/2018W01/P1.5Y"),
            IntervalResolution.Week,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R4/2018-306/P1.5Y"),
            IntervalResolution.Ordinal,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R5/2018W012/P1.5Y"),
            IntervalResolution.Weekday,
        )

    def test_get_interval_resolution_time(self):
        self.assertEqual(
            get_repeating_interval_resolution("R/P1M/1981-04-05T01"),
            IntervalResolution.Hours,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R1/P1M/1981-04-05T01:01"),
            IntervalResolution.Minutes,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R2/P1M/1981-04-05T01:01:00"),
            IntervalResolution.Seconds,
        )

        self.assertEqual(
            get_repeating_interval_resolution("R/1981-04-05T01/P1M"),
            IntervalResolution.Hours,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R1/1981-04-05T01:01/P1M"),
            IntervalResolution.Minutes,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R2/1981-04-05T01:01:00/P1M"),
            IntervalResolution.Seconds,
        )

    def test_get_interval_resolution_duration(self):
        self.assertEqual(
            get_repeating_interval_resolution("R/2014-11-12/P1Y2M3D"),
            IntervalResolution.Day,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R1/2014-11-12/P1Y2M"),
            IntervalResolution.Day,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R2/2014-11-12/P1Y"),
            IntervalResolution.Day,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R3/2014-11-12/P1W"),
            IntervalResolution.Day,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R4/2014-11-12/P1Y2M3DT4H"),
            IntervalResolution.Hours,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R5/2014-11-12/P1Y2M3DT4H54M"),
            IntervalResolution.Minutes,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R6/2014-11-12/P1Y2M3DT4H54M6S"),
            IntervalResolution.Seconds,
        )

        self.assertEqual(
            get_repeating_interval_resolution("R/P1Y2M3D/2014-11-12"),
            IntervalResolution.Day,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R1/P1Y2M/2014-11-12"),
            IntervalResolution.Day,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R2/P1Y/2014-11-12"),
            IntervalResolution.Day,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R3/P1W/2014-11-12"),
            IntervalResolution.Day,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R4/P1Y2M3DT4H/2014-11-12"),
            IntervalResolution.Hours,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R5/P1Y2M3DT4H54M/2014-11-12"),
            IntervalResolution.Minutes,
        )
        self.assertEqual(
            get_repeating_interval_resolution("R6/P1Y2M3DT4H54M6S/2014-11-12"),
            IntervalResolution.Seconds,
        )

    def test_parse_repeating_interval(self):
        with mock.patch.object(
            aniso8601.interval.PythonTimeBuilder, "build_repeating_interval"
        ) as mockBuilder:
            expectedargs = {
                "R": False,
                "Rnn": "3",
                "interval": IntervalTuple(
                    DateTuple("1981", "04", "05", None, None, None),
                    None,
                    DurationTuple(None, None, None, "1", None, None, None),
                ),
            }

            mockBuilder.return_value = expectedargs

            result = parse_repeating_interval("R3/1981-04-05/P1D")

            self.assertEqual(result, expectedargs)
            mockBuilder.assert_called_once_with(**expectedargs)

        with mock.patch.object(
            aniso8601.interval.PythonTimeBuilder, "build_repeating_interval"
        ) as mockBuilder:
            expectedargs = {
                "R": False,
                "Rnn": "11",
                "interval": IntervalTuple(
                    None,
                    DatetimeTuple(
                        DateTuple("1980", "03", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                    DurationTuple(None, None, None, None, "1", "2", None),
                ),
            }

            mockBuilder.return_value = expectedargs

            result = parse_repeating_interval("R11/PT1H2M/1980-03-05T01:01:00")

            self.assertEqual(result, expectedargs)
            mockBuilder.assert_called_once_with(**expectedargs)

        with mock.patch.object(
            aniso8601.interval.PythonTimeBuilder, "build_repeating_interval"
        ) as mockBuilder:
            expectedargs = {
                "R": False,
                "Rnn": "2",
                "interval": IntervalTuple(
                    DatetimeTuple(
                        DateTuple("1980", "03", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                    DatetimeTuple(
                        DateTuple("1981", "04", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                    None,
                ),
            }

            mockBuilder.return_value = expectedargs

            result = parse_repeating_interval(
                "R2--1980-03-05T01:01:00--" "1981-04-05T01:01:00",
                intervaldelimiter="--",
            )

            self.assertEqual(result, expectedargs)
            mockBuilder.assert_called_once_with(**expectedargs)

        with mock.patch.object(
            aniso8601.interval.PythonTimeBuilder, "build_repeating_interval"
        ) as mockBuilder:
            expectedargs = {
                "R": False,
                "Rnn": "2",
                "interval": IntervalTuple(
                    DatetimeTuple(
                        DateTuple("1980", "03", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                    DatetimeTuple(
                        DateTuple("1981", "04", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                    None,
                ),
            }

            mockBuilder.return_value = expectedargs

            result = parse_repeating_interval(
                "R2/" "1980-03-05 01:01:00/" "1981-04-05 01:01:00",
                datetimedelimiter=" ",
            )

            self.assertEqual(result, expectedargs)
            mockBuilder.assert_called_once_with(**expectedargs)

        with mock.patch.object(
            aniso8601.interval.PythonTimeBuilder, "build_repeating_interval"
        ) as mockBuilder:
            expectedargs = {
                "R": True,
                "Rnn": None,
                "interval": IntervalTuple(
                    None,
                    DatetimeTuple(
                        DateTuple("1980", "03", "05", None, None, None),
                        TimeTuple("01", "01", "00", None),
                    ),
                    DurationTuple(None, None, None, None, "1", "2", None),
                ),
            }

            mockBuilder.return_value = expectedargs

            result = parse_repeating_interval("R/PT1H2M/1980-03-05T01:01:00")

            self.assertEqual(result, expectedargs)
            mockBuilder.assert_called_once_with(**expectedargs)

    def test_parse_repeating_interval_mockbuilder(self):
        mockBuilder = mock.Mock()

        args = {
            "R": False,
            "Rnn": "3",
            "interval": IntervalTuple(
                DateTuple("1981", "04", "05", None, None, None),
                None,
                DurationTuple(None, None, None, "1", None, None, None),
            ),
        }

        mockBuilder.build_repeating_interval.return_value = args

        result = parse_repeating_interval("R3/1981-04-05/P1D", builder=mockBuilder)

        self.assertEqual(result, args)
        mockBuilder.build_repeating_interval.assert_called_once_with(**args)

        mockBuilder = mock.Mock()

        args = {
            "R": False,
            "Rnn": "11",
            "interval": IntervalTuple(
                None,
                DatetimeTuple(
                    DateTuple("1980", "03", "05", None, None, None),
                    TimeTuple("01", "01", "00", None),
                ),
                DurationTuple(None, None, None, None, "1", "2", None),
            ),
        }

        mockBuilder.build_repeating_interval.return_value = args

        result = parse_repeating_interval(
            "R11/PT1H2M/1980-03-05T01:01:00", builder=mockBuilder
        )

        self.assertEqual(result, args)
        mockBuilder.build_repeating_interval.assert_called_once_with(**args)

        mockBuilder = mock.Mock()

        args = {
            "R": True,
            "Rnn": None,
            "interval": IntervalTuple(
                None,
                DatetimeTuple(
                    DateTuple("1980", "03", "05", None, None, None),
                    TimeTuple("01", "01", "00", None),
                ),
                DurationTuple(None, None, None, None, "1", "2", None),
            ),
        }

        mockBuilder.build_repeating_interval.return_value = args

        result = parse_repeating_interval(
            "R/PT1H2M/1980-03-05T01:01:00", builder=mockBuilder
        )

        self.assertEqual(result, args)
        mockBuilder.build_repeating_interval.assert_called_once_with(**args)

    def test_parse_repeating_interval_badtype(self):
        testtuples = (None, 1, False, 1.234)

        for testtuple in testtuples:
            with self.assertRaises(ValueError):
                parse_repeating_interval(testtuple, builder=None)

    def test_parse_repeating_interval_baddelimiter(self):
        testtuples = ("R,PT1H2M,1980-03-05T01:01:00", "R3 1981-04-05 P1D")

        for testtuple in testtuples:
            with self.assertRaises(ISOFormatError):
                parse_repeating_interval(testtuple, builder=None)

    def test_parse_repeating_interval_suffixgarbage(self):
        # Don't allow garbage after the duration
        # https://bitbucket.org/nielsenb/aniso8601/issues/9/durations-with-trailing-garbage-are-parsed
        with self.assertRaises(ISOFormatError):
            parse_repeating_interval("R3/1981-04-05/P1Dasdf", builder=None)

        with self.assertRaises(ISOFormatError):
            parse_repeating_interval(
                "R3/" "1981-04-05/" "P0003-06-04T12:30:05.5asdfasdf", builder=None
            )

    def test_parse_repeating_interval_badstr(self):
        testtuples = ("bad", "")

        for testtuple in testtuples:
            with self.assertRaises(ISOFormatError):
                parse_repeating_interval(testtuple, builder=None)



import sys
from os import path

from setuptools import find_packages, setup

from aniso8601 import __version__

TESTS_REQUIRE = []

# Mock is only required for Python 2
PY2 = sys.version_info[0] == 2

if PY2:
    TESTS_REQUIRE.append("mock>=2.0.0")

THIS_DIRECTORY = path.abspath(path.dirname(__file__))
with open(path.join(THIS_DIRECTORY, "README.rst")) as f:
    README_TEXT = f.read()

setup(
    name="aniso8601",
    version=__version__,
    description="A library for parsing ISO 8601 strings.",
    long_description=README_TEXT,
    long_description_content_type="text/x-rst",
    author="Brandon Nielsen",
    author_email="nielsenb@jetfuse.net",
    url="https://codeberg.org/nielsenb-jf/aniso8601",
    packages=find_packages(),
    extras_require={
        "dev": TESTS_REQUIRE
        + [
            "black",
            "coverage",
            "isort",
            "pre-commit",
            "pyenchant",
            "pylint",
        ]
    },
    test_suite="aniso8601",
    tests_require=TESTS_REQUIRE,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="iso8601 parser",
    project_urls={
        "Changelog": "https://codeberg.org/nielsenb-jf/aniso8601/src/branch/main/CHANGELOG.rst",
        "Documentation": "https://aniso8601.readthedocs.io/",
        "Source": "https://codeberg.org/nielsenb-jf/aniso8601",
        "Tracker": "https://codeberg.org/nielsenb-jf/aniso8601/issues",
    },
)
